/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef IOHELPERS_HPP
#define IOHELPERS_HPP

#include <cstdarg>
#include <iostream>
#include <string>

inline std::string stringf (const std::string format, ...) {
	char buf[256];
	va_list args;
	va_start(args, format);
	vsnprintf(buf, 256, format.c_str(), args);
	va_end(args);
	return std::string(buf);
}

enum Yield {yield};
inline std::ostream &operator<< (std::ostream &out, const Yield &y) {out.flush(); throw y; return out;}

struct Expect {std::string tok;};
inline Expect expect (std::string tok) {Expect exp; exp.tok = tok; return exp;}
inline std::istream &operator>> (std::istream &in, Expect exp) {std::string tok; in >> tok; if (tok.compare(exp.tok)) std::cout << "Expected '" << exp.tok << "', found '" << tok << "' instead" << std::endl << yield; return in;}

inline std::string get_opt (const std::vector<char*> &args, const std::string &opt) {
	int nargs = (int) args.size();
	for (int i = 0; i < nargs; i++)
		if (opt.compare(args[i]) == 0)
			return args[i+1];
	return "";
}

inline int get_opti (const std::vector<char*> &args, const std::string &opt) {
	int nargs = (int) args.size();
	for (int i = 0; i < nargs; i++)
		if (opt.compare(args[i]) == 0)
			return atoi(args[i+1]);
	return (~0);
}

inline float get_optf (const std::vector<char*> &args, const std::string &opt) {
	int nargs = (int) args.size();
	for (int i = 0; i < nargs; i++)
		if (opt.compare(args[i]) == 0)
			return atof(args[i+1]);
	return not_a_number;
}

inline bool get_optb (const std::vector<char*> &args, const std::string &opt) {
	int nargs = (int) args.size();
	for (int i = 0; i < nargs; i++)
		if (opt.compare(args[i]) == 0)
			return true;
	return false;
}

#endif
