/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "marching.hpp"

using namespace std;

vector<Edge2D> extract_surface (const ScalarField<2> &rho, float rho_max, const ScalarField<2> &p) {
	int m = rho.m, n = rho.n;
	ScalarField<2> phi;
	phi.resize(m,n, rho.dx);
	for (int y = 0; y < n; y++) for (int x = 0; x < m; x++) {
			if (p(x+1,y+1) > 0) {
			phi(x,y) = 1;
			continue;
		}
		bool onring = false;
		float rho_min = rho(x,y);
		if (x > 0)   {onring |= p(x+0,y+1)>0; rho_min = fminf(rho_min, rho(x-1,y));}
		if (x < m-1) {onring |= p(x+2,y+1)>0; rho_min = fminf(rho_min, rho(x+1,y));}
		if (y > 0)   {onring |= p(x+1,y+0)>0; rho_min = fminf(rho_min, rho(x,y-1));}
		if (y < n-1) {onring |= p(x+1,y+2)>0; rho_min = fminf(rho_min, rho(x,y+1));}
		if (!onring) {
			phi(x,y) = -1;
			continue;
		}
		phi(x,y) = 2*(rho(x,y)-rho_min)/(rho_max-rho_min) - 1;
	}
	return marching_squares(phi);
}

#include <iostream>
vector<Edge2D> marching_squares (const ScalarField<2> &phi) {
	int m = phi.m, n = phi.n;
	float dx = phi.dx;
	vector<Edge2D> edges;
	for (int y = 1; y < n; y++) for (int x = 1; x < m; x++) {
		float phi0 = phi(x-1,y-1), phi1 = phi(x,y-1), phi2 = phi(x,y), phi3 = phi(x-1,y);
		int k = (phi0<0 ? 0 : 1) | (phi1<0 ? 0 : 2) | (phi2<0 ? 0 : 4) | (phi3<0 ? 0 : 8);
		if (k == 0 || k == 15) continue;
		float x0 = (x-1)*dx, y0 = (y-1)*dx, x1 = x*dx, y1 = y*dx;
		float px0 = phi0/(phi0-phi1), py0 = phi0/(phi0-phi3), px1 = phi3/(phi3-phi2), py1 = phi1/(phi1-phi2);
		//cout << phi0 << " " << phi1 << " " << phi2 << " " << phi3 << "\t";
		//cout << px0 << " " << py0 << " " << px1 << " " << py1 << endl;
		px0 = x0 + px0*dx; py0 = y0 + py0*dx; px1 = x0 + px1*dx; py1 = y0 + py1*dx;
		switch (k) {
		case 0:
		case 15:
			break;
		case 1:
		case 14:
			edges.push_back(Edge2D(Vec2f(px0,y0), Vec2f(x0,py0)));
			break;
		case 2:
		case 13:
			edges.push_back(Edge2D(Vec2f(px0,y0), Vec2f(x1,py1)));
			break;
		case 3:
		case 12:
			edges.push_back(Edge2D(Vec2f(x0,py0), Vec2f(x1,py1)));
			break;
		case 4:
		case 11:
			edges.push_back(Edge2D(Vec2f(px1,y1), Vec2f(x1,py1)));
			break;
		case 5:
		case 10:
			if ((phi0 + phi1 + phi2 + phi3 > 0) ^ (k!=5)) {
				edges.push_back(Edge2D(Vec2f(px0,y0), Vec2f(x1,py1)));
				edges.push_back(Edge2D(Vec2f(px1,y1), Vec2f(x0,py0)));
			} else {
				edges.push_back(Edge2D(Vec2f(px0,y0), Vec2f(x0,py0)));
				edges.push_back(Edge2D(Vec2f(px1,y1), Vec2f(x1,py1)));
			}
			break;
		case 6:
		case 9:
			edges.push_back(Edge2D(Vec2f(px0,y0), Vec2f(px1,y1)));
			break;
		case 7:
		case 8:
			edges.push_back(Edge2D(Vec2f(px1,y1), Vec2f(x0,py0)));
			break;
		}
	}
	return edges;
}
