/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "rihelpers.hpp"
#define _USE_MATH_DEFINES
#include <cmath>

using namespace std;

//inline float sqf(float x) {
//	return x * x;
//}

/// Renderman camera placement helper functions - Copied from The Renderman Companion (Listing 8.2, Pg 142/143)
void AimZ (RtPoint direction) {
	double xzlen, yzlen, yrot, xrot;

	if (direction[0] == 0 && direction[1] == 0 && direction[2] == 0) {
		return;
	}

	xzlen = sqrt (sqf(direction[0]) + sqf(direction[2]));
	if(xzlen == 0) 
		yrot = (direction[1] < 0) ? 180 : 0;
	else 
		yrot = 180 * acos (direction[2] / xzlen) / M_PI;

	yzlen = sqrt (sqf(direction[1]) + sqf(xzlen));
	xrot = 180 * acos (xzlen / yzlen) / M_PI; // yzlen shud not be 0

	if (direction[1] > 0)
		RiRotate (xrot, 1, 0, 0);
	else
		RiRotate (-xrot, 1, 0, 0);

	if (direction[0] > 0)
		RiRotate (-yrot, 0, 1, 0);
	else
		RiRotate (yrot, 0, 1, 0);
}

void PlaceCamera (RtPoint position, RtPoint direction, float roll, bool loadidentity) {
	if (loadidentity)
		RiIdentity();
	RiRotate (-roll, 0, 0, 1);
	AimZ (direction);
	RiTranslate (-position[0], -position[1], -position[2]);
}

void PlaceCameraPts (RtPoint from, RtPoint to, float roll, bool loadidentity) {
	RtPoint to2;
	to2[0] = to[0] - from[0];
	to2[1] = to[1] - from[1];
	to2[2] = to[2] - from[2];
	PlaceCamera (from, to2, roll, loadidentity);
}
