//#include "mex.h"
#include "../common.hpp"
//#include <assert.h>
#include "pcg_solver.h"

#ifndef __DOSTAL__
#define __DOSTAL__

void mexPrintVector (const char *name, const std::vector<float> &v);
void mexPrintVector (const char *name, const std::vector<double> &v);

enum stats_field {STEP, RESIDUAL, QUADRATIC, ACTIVE, TIME, NUM_STATS_FIELDS};

#define DEBUG 0
#define DEBUGVECTORS 0

#define mexPrintf printf

#if DEBUG
#define debugPrintf mexPrintf
template <typename T> inline void debugStats (std::vector<T> *stats, stats_field field, double x) {if (stats) stats[field].push_back(x);}
#else
inline void debugPrintf (const char *message, ...) {}
inline void debugStats (std::vector<float> *stats, stats_field field, double x) {}
#endif
#if DEBUGVECTORS
using namespace std;
template <typename T> inline void debugPrintVector (const char *name, const std::vector<T> &v) {
	cout << name << " = [";
	for (int i = 0; i < v.size(); i++) cout << (i==0?"":", ") << v[i];
	cout << "]" << endl;
}
#else
inline void debugPrintVector (const char *name, const std::vector<float> &v) {}
inline void debugPrintVector (const char *name, const std::vector<double> &v) {}
#endif

struct Dostal05Solver
{

public:

	SparseMatrix<float> A;
	std::vector<float> b, x;
	std::vector<float> xmin, xmax;
	float alpha_bar;

	// internal structures
	SparseColumnLowerFactor<float> M; // modified incomplete cholesky factor
	std::vector<float> r, p, z; // temporary vectors for PCG
	FixedSparseMatrix<float> A_fixed; // used within loop

	std::vector<float> pfg; // projected free gradient
	std::vector<float> a; // active set; 1 if at min, 0 if free, -1 if at max

	// parameters
	float tolerance;
	int max_iterations;
	float modified_incomplete_cholesky_parameter;
	float min_diagonal_ratio;

	float proportioning_factor;

public:

	Dostal05Solver(void);

	void set_solver_parameters(float tolerance_, int max_iterations_, float proportioning_factor_=1.0, float modified_incomplete_cholesky_parameter_=0.97, float min_diagonal_ratio_=0.25);

	void set_matrix(const SparseMatrix<float> &A);

	void set_rhs(const std::vector<float> &b);

	void set_initial(const std::vector<float> &x);

	void set_lower(const std::vector<float> &xmin);
	void set_upper(const std::vector<float> &xmax);

	bool solve(float &residual_initial_out, float &residual_out, int &iterations_out, std::vector<float> *stats=NULL);

};

#endif
