/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "qp.hpp"
#include "dostal05.h"

using namespace std;

void qp_solve_dostal (const SparseMatrix<float> &A, const vector<float> &b, vector<float> &x, QpReport &report) {
	Dostal05Solver solver;
	solver.set_solver_parameters(1e-3f, 100);
	solver.set_matrix(A);
	solver.set_rhs(b);
	solver.set_initial(x);
	report.success = solver.solve(report.residual_initial, report.residual, report.iterations, NULL);
	x = solver.x;
	if (!report.success) cout << "QP solver failed to converge; final residual = " << report.residual << endl;
}

#include <fstream>
void write_stats (ostream &out, vector<float> *stats) {
	for (int s = 0; s < NUM_STATS_FIELDS; s++) {
		switch (s) {
		case STEP: out << "step"; break;
		case RESIDUAL: out << "residual"; break;
		case QUADRATIC: out << "quadratic"; break;
		case ACTIVE: out << "active"; break;
		case TIME: out << "time"; break;
		}
		out << " = [ ";
		for (int i = 0; i < stats[s].size(); i++) {
			out << stats[s][i] << " ";
		}
		out << "];" << endl;
	}
}

void qp_solve_dostal (const SparseMatrix<float> &A, const vector<float> &b, const vector<float> &xmin, const vector<float> &xmax, vector<float> &x, QpReport &report) {
	Dostal05Solver solver;
	solver.set_solver_parameters(1e-3f, 100);
	solver.set_matrix(A);
	solver.set_rhs(b);
	solver.set_initial(x);
	solver.set_lower(xmin);
	solver.set_upper(xmax);
	//vector<float> stats[NUM_STATS_FIELDS];
	report.success = solver.solve(report.residual_initial, report.residual, report.iterations, NULL);
	/*fstream file("matlab/stats.m",ios::out);
	write_stats(file, stats);
	file.close();
	cin.ignore();*/
	x = solver.x;
	if (!report.success) cout << "QP solver failed to converge; final residual = " << report.residual << endl;
}
