#include "distance.hpp"

// lsNode.h
// Wed Jan  19 16:40 2005
// Copyright  2005 DNA Productions
// by Mark Carlson
#ifndef _LSNODE_H
#define _LSNODE_H
#define WIN_API //#include <WIN_API.h>

//#ifndef _FIELD3D_H
//#include <field3d.h>
//#endif
#include <vector>
typedef float REAL;
template <typename T> class field3d {
public:
	int I, J, K;
	std::vector<T> s;
	field3d &operator= (T val) {s.assign(I*J*K,val); return *this;}
	inline T& at (int i, int j, int k) {return s[i+I*(j+J*k)];}
	inline const T& at (int i, int j, int k) const {return s[i+I*(j+J*k)];}
	void resize (int I, int J, int K) {this->I = I; this->J = J; this->K = K; s.resize(I*J*K);}
};

#include <vector>

class WIN_API lsNode{
public:
  enum {locked=0,extend=-1,out=-2};
  short i,j,k,sn;
  REAL ls;
  lsNode(){}
  lsNode(int _i,int _j,int _k,REAL _ls){i=_i;j=_j;k=_k;ls=_ls;}//sn=ls<0?-1:1;}
  lsNode(int _i,int _j,int _k,REAL _ls,int _sn){i=_i;j=_j;k=_k;ls=_ls;sn=_sn;}//sn=ls<0?-1:1;}
  //friend bool operator<(const lsNode& a,const lsNode& b);//{return(a.ls>b.ls);}
};
inline bool operator<(const lsNode& a, const lsNode& b){return(a.ls>b.ls);}
// THis class is addapted from the heap class in Sedgwick's Algorithms in C++ 3rd ed

class WIN_API lsNode_heap{
        public:
                std::vector<lsNode> nodes;
          field3d<int>* field;// place to store the index into the heap of the nodes.
        //      int nClear,nEmpty,nPush,nPop,nChangeKeyS,nChangeKeyL,nMaxSize;
        public:
                // The nodes[0] is not used.
                lsNode_heap(field3d<int>& inF){field=&inF;nodes.push_back(lsNode(-1,-1,-1,0));
                        //nClear=nEmpty=nPush=nPop=nChangeKeyS=nChangeKeyL=nMaxSize=0;
                        }
                ~lsNode_heap(){nodes.clear();field=0;}
                
                void clear(){nodes.clear();nodes.push_back(lsNode(-1,-1,-1,0));}
                void reset(field3d<int>& inF){field=&inF;nodes.resize(0);nodes.push_back(lsNode(-1,-1,-1,0));}
                int empty(){return size()==0;}
                void push(const lsNode& item){
                //      nPush++;
	                /*field->s[item.i][item.j][item.k]*/field->at(item.i,item.j,item.k)=nodes.size();
                        nodes.push_back(item);
                        fixUp(size());
                //      if(nMaxSize<size())nMaxSize=size();
                }
                lsNode pop(){
        //              nPop++;
                        exch(1,size());
                        fixDown(1,size()-1);
                        nodes[0]=nodes.back();
                        nodes.pop_back();
                        /*field->s[nodes[0].i][nodes[0].j][nodes[0].k]*/field->at(nodes[0].i,nodes[0].j,nodes[0].k)=lsNode::locked;
                        return nodes[0];
                }
                void change_key(int i, int j, int k, REAL newKey){
	                int index=/*field->s[i][j][k]*/field->at(i,j,k);
                //      if(nodes[index].ls>=newKey)nChangeKeyS++;else nChangeKeyL++;
                        nodes[index].ls=newKey;
                        fixUp(index);
                        fixDown(index,size());
                }
                void change_key(int i, int j, int k, REAL newKey,short newSN){
	                int index=/*field->s[i][j][k]*/field->at(i,j,k);
                //      if(nodes[index].ls>=newKey)nChangeKeyS++;else nChangeKeyL++;
                        nodes[index].ls=newKey;
                        if(nodes[index].sn!=newSN){
                                nodes[index].ls=-newKey;
//                              printf("old<%i> new<%i> SIGN CHANGE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n",nodes[index].sn,newSN);
                        }
                        //nodes[index].sn=newSN;
                        fixUp(index);
                        fixDown(index,size());
                }
                size_t size(){return nodes.size()-1;}
        private:
                void fixUp(int k){
                        while(k>1 && nodes[k>>1] < nodes[k]){
                                exch(k,k>>1);
                                k>>=1;
                        }
                }
                void fixDown(int k, int N){
                        while((k<<1)<=N){
                                int j=k<<1;
                                if(j<N && nodes[j] < nodes[j+1])j++;
                                if(!(nodes[k] < nodes[j]))break;
                                exch(k,j);
                                k=j;
                        }
                }
                void exch(int a, int b){
	                /*field->s[nodes[a].i][nodes[a].j][nodes[a].k]*/field->at(nodes[a].i,nodes[a].j,nodes[a].k)=b;
	                /*field->s[nodes[b].i][nodes[b].j][nodes[b].k]*/field->at(nodes[b].i,nodes[b].j,nodes[b].k)=a;
                        nodes[0]=nodes[a];
                        nodes[a]=nodes[b];
                        nodes[b]=nodes[0];
                }
};

#endif // _LSNODE_H

/***************************************************************************
 *            vec.h
 *
 *  Tue Jan  4 15:05:15 2005
 *  Copyright  2005  Mark Carlson
 *  mcarlson@dnahelix.com
 ****************************************************************************/

#ifndef _VEC_H
#define _VEC_H
//#include <WIN_API.h>

#ifndef REAL
#define REAL float
#endif

#include <math.h>

class WIN_API vec{
public:
        REAL x,y,z;
        
        vec(){x=y=z=0;}//default constructor
        vec(const vec& inV){x=inV.x;y=inV.y;z=inV.z;}// copy constructor
        vec(REAL inX, REAL inY, REAL inZ){x=inX;y=inY;z=inZ;}// component constructor
        vec(REAL inF){x=y=z=inF;}// initial value constructor
        vec(float *fp){x=fp[0];y=fp[1];z=fp[2];}// initial value constructor
        
        vec& operator+=(const vec& inV){x+=inV.x;y+=inV.y;z+=inV.z;return(*this);}// component add assignment
        vec& operator-=(const vec& inV){x-=inV.x;y-=inV.y;z-=inV.z;return(*this);}// component subtract assignment
        vec& operator*=(REAL inF){x*=inF;y*=inF;z*=inF;return(*this);}// scalar multiply assignment
        vec& operator/=(REAL inF){x/=inF;y/=inF;z/=inF;return(*this);}// scalar divide assignment
        
        int operator! (){return((x||y||z)?0:1);} // logical negation.  1 if x=y=z=0, 0 otherwise
        vec operator- (){return(vec(-x, -y, -z));} // unary component negate
        vec operator- (const vec& inV){ return(vec(x-inV.x, y-inV.y, z-inV.z));} // component subtract
        vec operator+ (const vec& inV){ return(vec(x+inV.x, y+inV.y, z+inV.z)); } // component add
        vec operator* (REAL inScale){ return(vec(x*inScale, y*inScale, z*inScale)); } // vec * REAL
        vec operator/ (REAL inScale){ return(vec(x/inScale, y/inScale, z/inScale)); } // component divide
        REAL operator| (const vec& inV){ return(x*inV.x + y*inV.y + z*inV.z); } // dot product
        vec operator^ (const vec& inV){ return(vec(y*inV.z-z*inV.y, z*inV.x-x*inV.z, x*inV.y-y*inV.x));} // cross product
                
        // cast and pointer crap, assume float x,y,z; is stored like float vec[3];
        operator float*(){return &x;}
        float& operator[](int i){return (&x)[i];}
        /*bool operator< (vec& inV){
                if(x!=inV.x)return(x < inV.x);
    if(y!=inV.y)return(y < inV.y);
    return(z < inV.z);
        }                       
        bool operator== (vec& inV){
                return(x==inV.x && y==inV.y && z==inV.z);
        }               */

        REAL Length()const{return(sqrt(x*x + y*y + z*z));} // length of vector
        REAL LengthSquared()const{return(x*x + y*y + z*z);} // length of vector squared
        REAL Normalize() { REAL a(Length()); REAL b(a?(1.0f/a):0.0f); x *= b; y *= b; z *= b; return(a);} // normalize vector and return length
        REAL DistTo(const vec& inV){ return(sqrt((x-inV.x)*(x-inV.x)+(y-inV.y)*(y-inV.y)+(z-inV.z)*(z-inV.z)));} // distance between two points
        REAL DistToSquared(const vec& inV){ return((x-inV.x)*(x-inV.x)+(y-inV.y)*(y-inV.y)+(z-inV.z)*(z-inV.z));} // distance sqared between two points

        //friend vec operator* (REAL inScale, const vec& inV);
};

// 1.0 * foo
// inline gives internal linkage to non member functions
inline vec operator* (REAL inScale, const vec& inV) { return(vec(inV.x*inScale,inV.y*inScale,inV.z*inScale)); }
inline bool operator == (const vec& inV1, const vec& inV2){ return(((inV1.x == inV2.x)&&(inV1.y == inV2.y)&&(inV1.z == inV2.z))); }
inline bool operator != (const vec& inV1, const vec& inV2){ return(((inV1.x != inV2.x)||(inV1.y != inV2.y)||(inV1.z != inV2.z))); }
inline bool operator < (const vec& inV1, const vec& inV2){
  if(inV1.x!=inV2.x)return(inV1.x < inV2.x);
  if(inV1.y!=inV2.y)return(inV1.y < inV2.y);
  return(inV1.z < inV2.z); 
}

#endif /* _VEC_H */

/***************************************************************************
 *            SIM_LevelSet.h
 *
 *  Mon Mar 28 14:17:25 2005
 *  Copyright  2005  Mark Carlson
 *  mcarlson@dnahelix.com
 ****************************************************************************/

#ifndef _SIM_LEVELSET_H
#define _SIM_LEVELSET_H
#define WIN_API //#include <WIN_API.h>

//#include <GRID_Scalar.h>
class ScalarGrid : public field3d<float> {
public:
	ScalarGrid () {I = J = K = 0;}
	ScalarGrid (int I, int J, int K, float val) {this->I = I; this->J = J; this->K = K; s.assign(I*J*K,val);}
	float lirp(float x, float y, float z){
		// linear interpolation of a value in the field.  An out of range value will
		// simply project along its axis to nearest point (in that dimension) on the
		// field
		if(I==0||J==0||K==0)return 0.0f;
		int i=(int)x;
		int j=(int)y;
		int k=(int)z;
		float dx=x-i;
		float dy=y-j;
		float dz=z-k;
		// out of bounds check
		if(x<0.0f){i=0;dx=0.0f;}
		if(y<0.0f){j=0;dy=0.0f;}
		if(z<0.0f){k=0;dz=0.0f;}
		if(x>=I-1){i=I-2;dx=1.0f;}
		if(y>=J-1){j=J-2;dy=1.0f;}
		if(z>=K-1){k=K-2;dz=1.0f;}
		float lirpval =at(i  ,j  ,k  )*((1.f-dx)*(1.f-dy)*(1.f-dz));
		lirpval+=at(i  ,j  ,k+1)*((1.f-dx)*(1.f-dy)*dz      );
		lirpval+=at(i  ,j+1,k  )*((1.f-dx)*dy      *(1.f-dz));
		lirpval+=at(i  ,j+1,k+1)*((1.f-dx)*dy      *dz      );
		lirpval+=at(i+1,j  ,k  )*(dx      *(1.f-dy)*(1.f-dz));
		lirpval+=at(i+1,j  ,k+1)*(dx      *(1.f-dy)*dz      );
		lirpval+=at(i+1,j+1,k  )*(dx      *dy      *(1.f-dz));
		lirpval+=at(i+1,j+1,k+1)*(dx      *dy      *dz      );
		return lirpval;
	}
	float lirp (const vec &p) {return lirp(p.x,p.y,p.z);}
	vec lirpN(float x, float y, float z, bool useOldNormal){
		if(useOldNormal){
			float halfh=1.0f;
			float Nx = lirp(x+halfh,y,z) - lirp(x-halfh,y,z);
			float Ny = lirp(x,y+halfh,z) - lirp(x,y-halfh,z);
			float Nz = lirp(x,y,z+halfh) - lirp(x,y,z-halfh);
			float length=(float)sqrt(Nx*Nx+Ny*Ny+Nz*Nz);
			if(length)return(vec(Nx/length,Ny/length,Nz/length));
			else {// one sided dif
				Nx = lirp(x+halfh,y,z) - lirp(x,y,z);
				Ny = lirp(x,y+halfh,z) - lirp(x,y,z);
				Nz = lirp(x,y,z+halfh) - lirp(x,y,z);
				length=(float)sqrt(Nx*Nx+Ny*Ny+Nz*Nz);
				if(length)return(vec(Nx/length,Ny/length,Nz/length));
				else{
					Nx = lirp(x,y,z) - lirp(x-halfh,y,z);
					Ny = lirp(x,y,z) - lirp(x,y-halfh,z);
					Nz = lirp(x,y,z) - lirp(x,y,z-halfh);
					length=(float)sqrt(Nx*Nx+Ny*Ny+Nz*Nz);
					if(length)return(vec(Nx/length,Ny/length,Nz/length));
					else return(vec(0,0,0));
				}
			}
		} else {
			int i=(int)x;
			int j=(int)y;
			int k=(int)z;
			float dx=x-i;
			float dy=y-j;
			float dz=z-k;
			// out of bounds check
			if(x<0.0f){i=0;dx=0.0f;}
			if(y<0.0f){j=0;dy=0.0f;}
			if(z<0.0f){k=0;dz=0.0f;}
			if(x>=I-1){i=I-2;dx=1.0f;}
			if(y>=J-1){j=J-2;dy=1.0f;}
			if(z>=K-1){k=K-2;dz=1.0f;}
			float multiplier;
			vec lirpval(0,0,0);
			vec tempVec;
			// There are 8 corners that we may need, but if any of the
			// multipliers are 0 then don't bother wasting time getting
			// the normal for that node.
			if((multiplier=(1.f-dx)*(1.f-dy)*(1.f-dz))){
				tempVec=N(i  ,j  ,k  );
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=(1.f-dx)*(1.f-dy)*dz      )){
				tempVec=N(i  ,j  ,k+1);
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=(1.f-dx)*dy      *(1.f-dz))){
				tempVec=N(i  ,j+1,k  );
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=(1.f-dx)*dy      *dz      )){
				tempVec=N(i  ,j+1,k+1);
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=dx      *(1.f-dy)*(1.f-dz))){
				tempVec=N(i+1,j  ,k  );
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=dx      *(1.f-dy)*dz      )){
				tempVec=N(i+1,j  ,k+1);
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=dx      *dy      *(1.f-dz))){
				tempVec=N(i+1,j+1,k  );
				lirpval+=(multiplier*tempVec);
			}
			if((multiplier=dx      *dy      *dz      )){
				tempVec=N(i+1,j+1,k+1);
				lirpval+=(multiplier*tempVec);
			}
			lirpval.Normalize();
			return lirpval;
		}
	}
	vec lirpN (const vec &p, bool useOldNormal=true) {return lirpN(p.x,p.y,p.z,useOldNormal);}
	vec N(int i, int j, int k){
		// compute the normal at the grid point.  Use centered diffrencing unless
		// you are near the boundary. if |N|=0 by the end
		// then just return a 0 normal.
		float x,y,z;
		if(i<0)i=0;
		else if(i>I-1)i=I-1;
		if(j<0)j=0;
		else if(j>J-1)j=J-1;
		if(k<0)k=0;
		else if(k>K-1)k=K-1;
		// x=DCx(i,j,k);
		// y=DCy(i,j,k);
		// z=DCz(i,j,k);
		// compute the x component
		     if(i==0)x=at(i+1,j,k)-at(i,j,k);
		     else if(i==I-1)x=at(i,j,k)-at(i-1,j,k);
		     else x=0.5*(at(i+1,j,k)-at(i-1,j,k));
        
		// compute the y component
		     if(j==0)y=at(i,j+1,k)-at(i,j,k);
		     else if(j==J-1)y=at(i,j,k)-at(i,j-1,k);
		     else y=0.5*(at(i,j+1,k)-at(i,j-1,k));
        
		// compute the x component
		     if(k==0)z=at(i,j,k+1)-at(i,j,k);
		     else if(k==K-1)z=at(i,j,k)-at(i,j,k-1);
		     else z=0.5*(at(i,j,k+1)-at(i,j,k-1));
        
		vec returnMe(x,y,z);
		if(!returnMe);//do nothing, so we will be returning zero normals... LOOK OUT!
		else returnMe.Normalize();
        
		return(returnMe);
	}
};

//#include <GRID_Staggered.h>

class WIN_API LevelSet: public ScalarGrid {
public:
	LevelSet();// default constructor
	LevelSet(const LevelSet& LSin);// copy constructor
	LevelSet(int width, int height, int depth, int _minBandWidth, float val=0); // make a level set of size IJK and band with

	int minNarrowBandWidth;
	int narrowBandWidth;
	// dynamic arrays for level set bands
  std::vector<lsNode> bandSurface;
  std::vector<lsNode> bandPositive;
  std::vector<lsNode> bandNegative;
	std::vector<lsNode>* allBands[3];

	//void Advance(bool useBFECC, float dt, float dxInv, StaggeredGrid& velocity, vec& acceleration, ScalarGrid& bnd);
	//void AdvanceScalarBands(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids);
	//void ExtendScalarGridNegToPos(ScalarGrid& S, field3d<char>& extendThis);
	//void ExtendScalarGridPosToNeg(ScalarGrid& S, field3d<char>& extendThis);
	//void ExtendStaggeredGridNegToPos(StaggeredGrid& STAG, field3d<char>& extendThis);
	//void ExtendStaggeredGridPosToNeg(StaggeredGrid& STAG, field3d<char>& extendThis);
	void Reinitialize();
	//void SetNarrowBandWidth(float CFL);
	//int SaveSurface(char *fname, bool save_bands_only=false);
	//int LoadSurface(char *fname, field3d<char>* mask_p=0);
	//int LoadUnknown(char *fname, field3d<char>* mask_p=0);
	void RedistanceCrudelyPastBands(void);
	//void ClampDistancePastBands(void);	// added by VIVEK
	void InitBand(int BandWidth, bool haveSurface=false);
	void RedistanceBands(bool doPositive=true, bool doNegative=true, bool conditionalredistance=false);

private:
	//void AdvanceBFECC(float dt, float dxInv, StaggeredGrid& velocity);
	//void AdvanceSL(float dt, float dxInv, StaggeredGrid& velocity);
	//void AdvanceScalarBandsBFECC(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids);
	//void AdvanceScalarBandsSL(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids);
	void InitializeInterface(bool conditionalredistance=false, int maxIter=5, float tolerance=5e-6);
	void InitBand(){InitBand(narrowBandWidth);}
	//void RedistancePDE(int numSteps, float tau,bool haveSurface=false);
	float TentativeFMMVal(int i,int j,int k,float direction,field3d<int>& indexField,bool& failed);

};

#endif /* _SIM_LEVELSET_H */

#define myfmin std::min<float>
#define myfmax std::min<float>

/***************************************************************************
 *            SIM_LevelSet.cpp
 *
 *  Mon Mar 28 14:52:50 2005
 *  Copyright  2005  Mark Carlson
 *  mcarlson@dnahelix.com
 ****************************************************************************/
//#include <SIM_LevelSet.h>

//#ifndef _LSNODE_H
//#include <lsNode.h>
//#endif

#include <algorithm>
#include <iostream>
#include <queue>

LevelSet::LevelSet():ScalarGrid(){// default constructor
	minNarrowBandWidth=0;
	narrowBandWidth=0;
	allBands[0]=&bandSurface;
	allBands[1]=&bandPositive;
	allBands[2]=&bandNegative;
}  

LevelSet::LevelSet(const LevelSet& LSin):ScalarGrid(LSin){//copy constructor
	minNarrowBandWidth=LSin.minNarrowBandWidth;
	narrowBandWidth=LSin.narrowBandWidth;
	bandSurface=LSin.bandSurface;
	bandPositive=LSin.bandPositive;
	bandNegative=LSin.bandNegative;
	allBands[0]=&bandSurface;
	allBands[1]=&bandPositive;
	allBands[2]=&bandNegative;
}
LevelSet::LevelSet(int width, int height, int depth, int _minBandWidth, float val):ScalarGrid(width,height,depth,val){
  bandSurface.clear();
	allBands[0]=&bandSurface;
	allBands[1]=&bandPositive;
	allBands[2]=&bandNegative;
	minNarrowBandWidth=_minBandWidth;
	narrowBandWidth=minNarrowBandWidth;
}

// void LevelSet::Advance(bool useBFECC,float dt, float dxInv, StaggeredGrid& velocity, vec& acceleration, ScalarGrid& bnd){
// 	if(useBFECC)AdvanceBFECC(dt, dxInv, velocity);
// 	else AdvanceSL(dt, dxInv, velocity);
// }

// void LevelSet::AdvanceScalarBands(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids){
// 	AdvanceScalarBandsBFECC(dt,dxInv,velocity,grids);
// 	//AdvanceScalarBandsSL(dt,dxInv,velocity,grids);
// }

// /* LevelSet::ExtendScalarGridNegToPos-- extend a scalar quantity, along the normal
// //
// // extendThis should be set to lsNode::locked for cells with valid values
// // extendThis should be set to lsNode::extend for all nodes that wish to be extended.
// // extendThis should be set to lsNode::out for all nodes that are ignored
// // 
// // cells outside the range of the grid are considered ::out
// */
// void LevelSet::ExtendScalarGridNegToPos(ScalarGrid& S, field3d<char>& extendThis){
// 	std::vector<lsNode> nodes;
// 	int numFailed=0;

// 	//osherFedkiw p.75 for normal information. p.77 for derivation.
// 	for(int m=0;m<2;m++){//surface and air band cells
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			if(extendThis.s[i][j][k]==lsNode::extend){
// 				nodes.push_back(lsNode(i,j,k,s[i][j][k]));
// 			}
// 		}
// 	}
// 	std::sort(nodes.begin(),nodes.end());// sorts in reverse of what we need.	
// //	printf("low to high 2");
// 	for(int n=nodes.size()-1;n>=0;n--){//compute the extended Positive
// //		printf("%g ", nodes[n].ls);
// 		int i=nodes[n].i;
// 		int j=nodes[n].j;
// 		int k=nodes[n].k;
// 		// compute the normal so information only flows from low to higher
// 		// first find the levelset value at the point we are at lsM, then for the
// 		// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 		// or lsL is < lsM then use the minimim to compute the one sided difference.
// 		// If both are larger, then we set LS_x=0 because no velocity information
// 		// should be flowing into that point from the x direction.
// 		// do similar for the LS_y, and LS_z
// 		REAL lsM=s[i][j][k];
// 		REAL lsL=i?s[i-1][j][k]:lirp(i-1,j,k);
// 		REAL lsR=(i<I-1)?s[i+1][j][k]:lirp(i+1,j,k);
// 		REAL LS_x;
// 		if(myfmin(lsL,lsR)>=lsM){
// 			LS_x=0.0;
// 		} else if(lsL<lsR){
// 			LS_x=lsM-lsL;
// 		} else LS_x=lsR-lsM;
			
// 		lsL=j?s[i][j-1][k]:lirp(i,j-1,k);
// 		lsR=(j<J-1)?s[i][j+1][k]:lirp(i,j+1,k);
		
// 		REAL LS_y;
// 		if(myfmin(lsL,lsR)>=lsM){
// 			LS_y=0.0;
// 		} else if(lsL<lsR){
// 			LS_y=lsM-lsL;
// 		} else LS_y=lsR-lsM;
		
// 		lsL=k?s[i][j][k-1]:lirp(i,j,k-1);
// 		lsR=(k<K-1)?s[i][j][k+1]:lirp(i,j,k+1);
		
// 		REAL LS_z;
// 		if(myfmin(lsL,lsR)>=lsM){
// 			LS_z=0.0;
// 		} else if(lsL<lsR){
// 			LS_z=lsM-lsL;
// 		} else LS_z=lsR-lsM;
// 		// Now we solve the equation LS_x*S_x + LS_y*S_y + LS_z*S_z = 0
// 		// we use upwind differencing to solve the scalar field derivatives
// 		// and the LS derivatives are what we use to decide the upwind direction.
// 		// So.  If LS_x >0 we use S_x=(S[i][j][k]-S[i-1][j][k])/dx
// 		if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 			REAL Numerator=0;
// 			REAL Denominator=0;
// 			if(LS_x){
// 				Numerator+=LS_x>0?S.s[i-1][j][k]*LS_x:-S.s[i+1][j][k]*LS_x;
// 				Denominator+=LS_x>0?LS_x:-LS_x;
// 			}
// 			if(LS_y){
// 				Numerator+=LS_y>0?S.s[i][j-1][k]*LS_y:-S.s[i][j+1][k]*LS_y;
// 				Denominator+=LS_y>0?LS_y:-LS_y;
// 			}
// 			if(LS_z){
// 				Numerator+=LS_z>0?S.s[i][j][k-1]*LS_z:-S.s[i][j][k+1]*LS_z;
// 				Denominator+=LS_z>0?LS_z:-LS_z;
// 			}
// 			S.s[i][j][k]=Numerator/Denominator;
// 		} else {
// 			numFailed++;
// 		}
// 	}
// 	#ifdef _DEBUG
// 	if(numFailed){
// 		printf("ExtendScalarGridNegToPos failed:");
// 		if(numFailed)printf(" <%i>",numFailed);
// 		printf("\n");
// 	}
// 	#endif
// }

// /* LevelSet::ExtendScalarGridPosToNeg-- extend a scalar quantity, along the negative normal
// //
// // extendThis should be set to lsNode::locked for cells with valid values
// // extendThis should be set to lsNode::extend for all nodes that wish to be extended.
// // extendThis should be set to lsNode::out for all nodes that are ignored
// // 
// // cells outside the range of the grid are considered ::out
// */
// void LevelSet::ExtendScalarGridPosToNeg(ScalarGrid& S, field3d<char>& extendThis){
// 	std::vector<lsNode> nodes;
// 	int numFailed=0;
	
// 	//osherFedkiw p.75 for normal information. p.77 for derivation.
// 	for(int m=0;m<2;m++){//surface and air band cells
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			if(extendThis.s[i][j][k]==lsNode::extend){
// 				nodes.push_back(lsNode(i,j,k,s[i][j][k]));
// 			}
// 		}
// 	}
// 	std::sort(nodes.begin(),nodes.end());// sorts in exact order we need.
// 	for(unsigned int n=0;n<nodes.size();n++){//compute the extended negative
// 		int i=nodes[n].i;
// 		int j=nodes[n].j;
// 		int k=nodes[n].k;
// 		// compute the normal so information only flows from higher to lower
// 		// first find the boundary levelset value at the point we are at lsM, then for the
// 		// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 		// or lsL is > lsM then use the maximum to compute the one sided difference.
// 		// If both are smaller, then we set LS_x=0 because no velocity information
// 		// should be flowing into that point from the x direction.
// 		// do similar for the LS_y, and LS_z
// 		REAL lsM=s[i][j][k];
// 		REAL lsL=i?s[i-1][j][k]:lirp(i-1,j,k);
// 		REAL lsR=(i<I-1)?s[i+1][j][k]:lirp(i+1,j,k);
		
// 		REAL LS_x;
// 		if(myfmax(lsL,lsR)<=lsM){
// 			LS_x=0.0;
// 		} else if(lsL>lsR){
// 			LS_x=lsM-lsL;
// 		} else LS_x=lsR-lsM;
			
// 		lsL=j?s[i][j-1][k]:lirp(i,j-1,k);
// 		lsR=(j<J-1)?s[i][j+1][k]:lirp(i,j+1,k);

// 		REAL LS_y;
// 		if(myfmax(lsL,lsR)<=lsM){
// 			LS_y=0.0;
// 		} else if(lsL>lsR){
// 			LS_y=lsM-lsL;
// 		} else LS_y=lsR-lsM;
			
// 		lsL=k?s[i][j][k-1]:lirp(i,j,k-1);
// 		lsR=(k<K-1)?s[i][j][k+1]:lirp(i,j,k+1);

// 		REAL LS_z;
// 		if(myfmax(lsL,lsR)<=lsM){
// 			LS_z=0.0;
// 		} else if(lsL>lsR){
// 			LS_z=lsM-lsL;
// 		} else LS_z=lsR-lsM;
// 		// Now we solve the equation LS_x*S_x + LS_y*S_y + LS_z*S_z = 0
// 		// we use upwind differencing to solve the velocity derivatives
// 		// and the LS derivatives are what we use to decide the upwind direction.
// 		// So.  If LS_x >0 we use S_x=(S[i][j][k]-S[i-1][j][k])/dx
// 		if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 			REAL Numerator=0;
// 			REAL Denominator=0;
// 			if(LS_x){
// 				Numerator+=LS_x<0?
// 				  S.s[i-1][j][k]*LS_x:
// 				  -S.s[i+1][j][k]*LS_x;
// 				Denominator+=LS_x<0?LS_x:-LS_x;
// 			}
// 			if(LS_y){
// 				Numerator+=LS_y<0?S.s[i][j-1][k]*LS_y:-S.s[i][j+1][k]*LS_y;
// 				Denominator+=LS_y<0?LS_y:-LS_y;
// 			}
// 			if(LS_z){
// 				Numerator+=LS_z<0?S.s[i][j][k-1]*LS_z:-S.s[i][j][k+1]*LS_z;
// 				Denominator+=LS_z<0?LS_z:-LS_z;
// 			}
// 			S.s[i][j][k]=Numerator/Denominator;
// 		} else {
// 			numFailed++;
// 		}
// 	}
// 	#ifdef _DEBUG
// 	if(numFailed){
// 		printf("ExtendScalarGridPosToNeg failed:");
// 		if(numFailed)printf(" <%i>",numFailed);
// 		printf("\n");
// 	}
// 	#endif
// }

// /* LevelSet::ExtendStagGridNegToPos-- extend a vector quantity, stored on a staggared grid, along the normal of a level set
// //
// // extendThis should be set to lsNode::locked for cells with valid velocities on all faces,
// //  but if one face borders an extend cell then that face will have its value over written
// // extendThis should be set to lsNode::extend for all nodes that wish to be extended on all faces.
// //  but if one face borders an out cell, then that face will not be extended.
// // extendThis should be set to lsNode::out for all nodes that are at don't want any of their their faces extenede
// // 
// // cells outside the range of the grid are considered ::out
// */
// void foo(void){;} // take out this foo that does nothing and you will get a crash... wtf !!! use -ffloat-store in CFLAGS
// void LevelSet::ExtendStaggeredGridNegToPos(StaggeredGrid& STAG, field3d<char>& extendThis){
// 	std::vector<lsNode> nodes;
// 	int numFailedU=0;
// 	int numFailedV=0;
// 	int numFailedW=0;
	
// 	// In order to do this in a for loop instead of code replication for each
// 	// velocity, we must creat 3 arrays.
// 	field3d<REAL>* VEL[3]={&(STAG.X),&(STAG.Y),&(STAG.Z)};// the current velocity being calculated.
// 	const int ngh[3][3]={{-1,0,0},{0,-1,0},{0,0,-1}};// offset location of the NeiGHbor cell that shares the face with cell[i][j][k]
// 	const float loc[3][3]={{-0.5,0,0},{0,-0.5,0},{0,0,-0.5}};// offset from i,j,k of the node LOCation in the grid
// 	int ijk[3];// will hold i,j,k
// 	int IJK[3]={I,J,K};

// 	//osherFedkiw p.75 for normal information. p.77 for derivation.
// 	//trouble is that that book uses centered velocities, not face velocities
// 	// Solve for U, remember for cell i,j,k, the velocity is at i-0.5,j,k
// 	for(int dim=0;dim<3;dim++){
// 		for(int m=0;m<2;m++){//surface and air band cells
// 			std::vector<lsNode>& curBand=*(allBands[m]);
// 			for(unsigned int n=0;n<curBand.size();n++){
// 				int i=curBand[n].i;
// 				int j=curBand[n].j;
// 				int k=curBand[n].k;
// 				if(extendThis.s[i][j][k]==lsNode::extend){
// 					ijk[0]=i;
// 					ijk[1]=j;
// 					ijk[2]=k;
// 					int iLNgh=i+ngh[dim][0];// left neighbor
// 					int jLNgh=j+ngh[dim][1];
// 					int kLNgh=k+ngh[dim][2];
// 					int iRNgh=i-ngh[dim][0];// right neighbor
// 					int jRNgh=j-ngh[dim][1];
// 					int kRNgh=k-ngh[dim][2];
// 					REAL lsVal=s[i][j][k];// simply initialize to something to get rid or compiler warning
// 					// left face
// 					if(ijk[dim] && !(extendThis.s[iLNgh][jLNgh][kLNgh]==lsNode::out)){
// 						lsVal=0.5*(s[i][j][k]+s[iLNgh][jLNgh][kLNgh]);
// 						nodes.push_back(lsNode(i,j,k,lsVal));
// 					}
// 					// right face (rare)
// 					if((!(ijk[dim]==IJK[dim]-1)) && extendThis.s[iRNgh][jRNgh][kRNgh]==lsNode::locked){
// 						lsVal=0.5*(s[i][j][k]+s[iRNgh][jRNgh][kRNgh]);
// 						nodes.push_back(lsNode(iRNgh,jRNgh,kRNgh,lsVal));
// 					}
// 				}
// 			}
// 		}
		
// 		std::sort(nodes.begin(),nodes.end());// sorts in reverse of what we need.	
// //		printf("low to high ");

// //		int A1,A2,A3,B1,B2,B3,C1,C2,C3,D1,D2,D3,D4;
// //		A1=A2=A3=B1=B2=B3=C1=C2=C3=D1=D2=D3=D4=0;
// 		for(int n=nodes.size()-1;n>=0;n--){//compute the extended Positive
// 			//printf("n=%i\n",n);fflush(0);//119746
// 			int i=nodes[n].i;
// 			int j=nodes[n].j;
// 			int k=nodes[n].k;
// 			REAL iLoc=i+loc[dim][0];
// 			REAL jLoc=j+loc[dim][1];
// 			REAL kLoc=k+loc[dim][2];
// 			// compute the normal so information only flows from low to higher
// 			// first find the levelset value at the point we are at lsM, then for the
// 			// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 			// or lsL is < lsM then use the minimim to compute the one sided difference.
// 			// If both are larger, then we set LS_x=0 because no velocity information
// 			// should be flowing into that point from the x direction.
// 			// do similar for the LS_y, and LS_z
// 			REAL lsM=lirp(iLoc,jLoc,kLoc);
// 			REAL lsL=lirp(iLoc-1,jLoc,kLoc);
// 			REAL lsR=lirp(iLoc+1,jLoc,kLoc);
// 			//foo();
// 			//REAL LS_x=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_x;
// 			if(((float)myfmin(lsL,lsR))>=lsM){
// //				A1++;
// 				LS_x=0.0;
// 			} else if(lsL<lsR){
// //				A2++;
// 				LS_x=lsM-lsL;
// 			} else {
// //				A3++;
// 				LS_x=lsR-lsM;
// 			}
// 			lsL=lirp(iLoc,jLoc-1,kLoc);
// 			lsR=lirp(iLoc,jLoc+1,kLoc);
// 			//REAL LS_y=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_y;
// 			if(((float)myfmin(lsL,lsR))>=lsM){
// //				B1++;
// 				LS_y=0.0;
// 			} else if(lsL<lsR){
// //				B2++;
// 				LS_y=lsM-lsL;
// 			} else {
// //				B3++;
// 				LS_y=lsR-lsM;
// 			}
// 			lsL=lirp(iLoc,jLoc,kLoc-1);
// 			lsR=lirp(iLoc,jLoc,kLoc+1);
// 			//REAL LS_z=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_z;
// 			if(((float)myfmin(lsL,lsR))>=lsM){
// //				C1++;
// 				LS_z=0.0;
// 			} else if(lsL<lsR){
// //				C2++;
// 				LS_z=lsM-lsL;
// 			} else {
// //				C3++;
// 				LS_z=lsR-lsM;
// 			}
// 			// Now we solve the equation LS_x*U_x + LS_y*U_y + LS_z*U_z = 0
// 			// we use upwind differencing to solve the velocity derivatives
// 			// and the LS derivatives are what we use to decide the upwind direction.
// 			// So.  If LS_x >0 we use U_x=(U[i][j][k]-U[i-1][j][k])/dx
// 			if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 				REAL Numerator=0;
// 				REAL Denominator=0;
// 				if(LS_x){
// //					D1++;
// 					Numerator+=LS_x>0?VEL[dim]->s[i-1][j][k]*LS_x:-VEL[dim]->s[i+1][j][k]*LS_x;
// 					Denominator+=LS_x>0?LS_x:-LS_x;
// 				}
// 				if(LS_y){
// //					D2++;
// 					Numerator+=LS_y>0?VEL[dim]->s[i][j-1][k]*LS_y:-VEL[dim]->s[i][j+1][k]*LS_y;
// 					Denominator+=LS_y>0?LS_y:-LS_y;
// 				}
// 				if(LS_z){
// //					D3++;
// 					Numerator+=LS_z>0?VEL[dim]->s[i][j][k-1]*LS_z:-VEL[dim]->s[i][j][k+1]*LS_z;
// 					Denominator+=LS_z>0?LS_z:-LS_z;
// 				}
// 				VEL[dim]->s[i][j][k]=Numerator/Denominator;
// 			} else {
// //				D4++;
				
// 				dim?dim==1?numFailedV++:numFailedW++:numFailedU++;
// 			}
// 		}
// //		printf("DIM<%i>\n",dim);
// //		printf("A1=%i\nA2=%i\nA3=%i\n\n",A1,A2,A3);fflush(0);
// //		printf("B1=%i\nB2=%i\nB3=%i\n\n",B1,B2,B3);fflush(0);
// //		printf("C1=%i\nC2=%i\nC3=%i\n\n",C1,C2,C3);fflush(0);
// //		printf("D1=%i\nD2=%i\nD3=%i\nD4=%i\n\n",D1,D2,D3,D4);fflush(0);
// 		nodes.clear();
// 	}
// 	#ifdef _DEBUG
// 	if(numFailedU||numFailedV||numFailedW){
// 		printf("ExtendStagGridNegToPos failed:");
// 		if(numFailedU)printf(" U<%i>",numFailedU);
// 		if(numFailedV)printf(" V<%i>",numFailedV);
// 		if(numFailedW)printf(" W<%i>",numFailedW);
// 		printf("\n");fflush(0);
// 	}
// 	#endif
// }
// /* LevelSet::ExtendStagGridPosToNeg-- extend a vector quantity, stored on a staggared grid, along the normal of a level set
// //
// // extendThis should be set to lsNode::locked for cells with valid velocities on all faces,
// //  but if one face borders an extend cell then that face will have its value over written
// // extendThis should be set to lsNode::extend for all nodes that wish to be extended on all faces.
// //  but if one face borders an out cell, then that face will not be extended.
// // extendThis should be set to lsNode::out for all nodes that are at don't want any of their their faces extenede
// // 
// // cells outside the range of the grid are considered ::out
// */
// void LevelSet::ExtendStaggeredGridPosToNeg(StaggeredGrid& STAG, field3d<char>& extendThis){
// 	std::vector<lsNode> nodes;
// 	int numFailedU=0;
// 	int numFailedV=0;
// 	int numFailedW=0;
	
// 	// In order to do this in a for loop instead of code replication for each
// 	// velocity, we must creat 3 arrays.
// 	field3d<REAL>* VEL[3]={&(STAG.X),&(STAG.Y),&(STAG.Z)};// the current velocity being calculated.
// 	const int ngh[3][3]={{-1,0,0},{0,-1,0},{0,0,-1}};// offset location of the NeiGHbor cell that shares the face with cell[i][j][k]
// 	const float loc[3][3]={{-0.5,0,0},{0,-0.5,0},{0,0,-0.5}};// offset from i,j,k of the node LOCation in the grid
// 	int ijk[3];// will hold i,j,k
// 	int IJK[3]={I,J,K};

// 	//osherFedkiw p.75 for normal information. p.77 for derivation.
// 	//trouble is that that book uses centered velocities, not face velocities
// 	// Solve for U, remember for cell i,j,k, the velocity is at i-0.5,j,k
// 	for(int dim=0;dim<3;dim++){
// 		for(int m=0;m<2;m++){//surface and air band cells
// 			std::vector<lsNode>& curBand=*(allBands[m]);
// 			for(unsigned int n=0;n<curBand.size();n++){
// 				int i=curBand[n].i;
// 				int j=curBand[n].j;
// 				int k=curBand[n].k;
// 				if(extendThis.s[i][j][k]==lsNode::extend){
// 					ijk[0]=i;
// 					ijk[1]=j;
// 					ijk[2]=k;
// 					int iLNgh=i+ngh[dim][0];// left neighbor
// 					int jLNgh=j+ngh[dim][1];
// 					int kLNgh=k+ngh[dim][2];
// 					int iRNgh=i-ngh[dim][0];// right neighbor
// 					int jRNgh=j-ngh[dim][1];
// 					int kRNgh=k-ngh[dim][2];
// 					REAL lsVal=s[i][j][k];// simply initialize to something to get rid or compiler warning
// 					// left face
// 					if(ijk[dim] && !(extendThis.s[iLNgh][jLNgh][kLNgh]==lsNode::out)){
// 						lsVal=0.5*(s[i][j][k]+s[iLNgh][jLNgh][kLNgh]);
// 						nodes.push_back(lsNode(i,j,k,lsVal));
// 					}
// 					// right face (rare)
// 					if((!(ijk[dim]==IJK[dim]-1)) && extendThis.s[iRNgh][jRNgh][kRNgh]==lsNode::locked){
// 						lsVal=0.5*(s[i][j][k]+s[iRNgh][jRNgh][kRNgh]);
// 						nodes.push_back(lsNode(iRNgh,jRNgh,kRNgh,lsVal));
// 					}
// 				}
// 			}
// 		}
// 		std::sort(nodes.begin(),nodes.end());// sorts in exact order we need.
// 		for(unsigned int n=0;n<nodes.size();n++){//compute the extended negative
// 			int i=nodes[n].i;
// 			int j=nodes[n].j;
// 			int k=nodes[n].k;
// 			REAL iLoc=i+loc[dim][0];
// 			REAL jLoc=j+loc[dim][1];
// 			REAL kLoc=k+loc[dim][2];
// 			// compute the normal so information only flows from higher to lower
// 			// first find the boundary levelset value at the point we are at lsM, then for the
// 			// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 			// or lsL is > lsM then use the maximum to compute the one sided difference.
// 			// If both are smaller, then we set LS_x=0 because no velocity information
// 			// should be flowing into that point from the x direction.
// 			// do similar for the LS_y, and LS_z
// 			REAL lsM=lirp(iLoc,jLoc,kLoc);
// 			REAL lsL=lirp(iLoc-1,jLoc,kLoc);
// 			REAL lsR=lirp(iLoc+1,jLoc,kLoc);
// 			foo();
// 			//REAL LS_x=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_x;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_x=0.0;
// 			} else if(lsL>lsR){
// 				LS_x=lsM-lsL;
// 			} else LS_x=lsR-lsM;
			
// 			lsL=lirp(iLoc,jLoc-1,kLoc);
// 			lsR=lirp(iLoc,jLoc+1,kLoc);
// 			//REAL LS_y=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_y;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_y=0.0;
// 			} else if(lsL>lsR){
// 				LS_y=lsM-lsL;
// 			} else LS_y=lsR-lsM;
			
// 			lsL=lirp(iLoc,jLoc,kLoc-1);
// 			lsR=lirp(iLoc,jLoc,kLoc+1);
// 			//REAL LS_z=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_z;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_z=0.0;
// 			} else if(lsL>lsR){
// 				LS_z=lsM-lsL;
// 			} else LS_z=lsR-lsM;
// 			// Now we solve the equation LS_x*U_x + LS_y*U_y + LS_z*U_z = 0
// 			// we use upwind differencing to solve the velocity derivatives
// 			// and the LS derivatives are what we use to decide the upwind direction.
// 			// So.  If LS_x >0 we use U_x=(U[i][j][k]-U[i-1][j][k])/dx
// 			if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 				REAL Numerator=0;
// 				REAL Denominator=0;
// 				if(LS_x){
// 					Numerator+=LS_x<0?
// 					  VEL[dim]->s[i-1][j][k]*LS_x:
// 					  -VEL[dim]->s[i+1][j][k]*LS_x;
// 					Denominator+=LS_x<0?LS_x:-LS_x;
// 				}
// 				if(LS_y){
// 					Numerator+=LS_y<0?VEL[dim]->s[i][j-1][k]*LS_y:-VEL[dim]->s[i][j+1][k]*LS_y;
// 					Denominator+=LS_y<0?LS_y:-LS_y;
// 				}
// 				if(LS_z){
// 					Numerator+=LS_z<0?VEL[dim]->s[i][j][k-1]*LS_z:-VEL[dim]->s[i][j][k+1]*LS_z;
// 					Denominator+=LS_z<0?LS_z:-LS_z;
// 				}
// 				VEL[dim]->s[i][j][k]=Numerator/Denominator;
// 			} else {
// 				dim?dim==1?numFailedV++:numFailedW++:numFailedU++;
// 			}
// 		}
// 		nodes.clear();
// 	}
// 	#ifdef _DEBUG
// 	if(numFailedU||numFailedV||numFailedW){
// 		printf("ExtendStagGridPosToNeg failed:");
// 		if(numFailedU)printf(" U<%i>",numFailedU);
// 		if(numFailedV)printf(" V<%i>",numFailedV);
// 		if(numFailedW)printf(" W<%i>",numFailedW);
// 		printf("\n");
// 	}
// 	#endif
// }
// /* LevelSet::ExtendStagGrid-- extend a vector quantity, stored on a staggared grid, along the normal of a level set
// //
// // extendThis should be set to lsNode::locked for cells with valid velocities on all faces,
// //  but if one face borders an extend cell then that face will have its value over written
// // extendThis should be set to lsNode::extend for all nodes that wish to be extended on all faces.
// //  but if one face borders an out cell, then that face will not be extended.
// // extendThis should be set to lsNode::out for all nodes that are at don't want any of their their faces extenede
// //
// // All lsNode::extend nodes will be sorted into two groups, one for positive ls and one for negative ls.
// // the positive ones will be extended from smallest to largest along the ls normal.
// // the negative ones will be extended from largest to smallest along the negative ls normal.
// // 
// // cells outside the range of the grid are considered ::out
// *//*
// void LevelSet::ExtendStagGrid(StaggeredGrid& STAG, field3d<char>& extendThis){
// 	std::vector<lsNode> positiveNodes;
// 	std::vector<lsNode> negativeNodes;
// 	int numFailedUn=0;
// 	int numFailedVn=0;
// 	int numFailedWn=0;
// 	int numFailedUp=0;
// 	int numFailedVp=0;
// 	int numFailedWp=0;
	
// 	// In order to do this in a for loop instead of code replication for each
// 	// velocity, we must creat 3 arrays.
// 	field3d<REAL>* VEL[3]={&(STAG.X),&(STAG.Y),&(STAG.Z)};// the current velocity being calculated.
// 	const int ngh[3][3]={{-1,0,0},{0,-1,0},{0,0,-1}};// offset location of the NeiGHbor cell that shares the face with cell[i][j][k]
// 	const float loc[3][3]={{-0.5,0,0},{0,-0.5,0},{0,0,-0.5}};// offset from i,j,k of the node LOCation in the grid
// 	int ijk[3];// will hold i,j,k
// 	int IJK[3]={I,J,K};

// 	//osherFedkiw p.75 for normal information. p.77 for derivation.
// 	//trouble is that that book uses centered velocities, not face velocities
// 	// Solve for U, remember for cell i,j,k, the velocity is at i-0.5,j,k
// 	for(int dim=0;dim<3;dim++){
// 		// push in the nodes that need to be extended.
// 		for(int i=0;i<I;i++)for(int j=0;j<J;j++)for(int k=0;k<K;k++){
// 			if(extendThis.s[i][j][k]==lsNode::extend){
// 				ijk[0]=i;
// 				ijk[1]=j;
// 				ijk[2]=k;
// 				int iLNgh=i+ngh[dim][0];// left neighbor
// 				int jLNgh=j+ngh[dim][1];
// 				int kLNgh=k+ngh[dim][2];
// 				int iRNgh=i-ngh[dim][0];// right neighbor
// 				int jRNgh=j-ngh[dim][1];
// 				int kRNgh=k-ngh[dim][2];
// 				REAL lsVal=s[i][j][k];// simply initialize to something to get rid or compiler warning
// 				// left face
// 				bool extendFace=false;
// 				if(ijk[dim] && !(extendThis.s[iLNgh][jLNgh][kLNgh]==lsNode::out)){
// 					extendFace=true;
// 					lsVal=0.5*(s[i][j][k]+s[iLNgh][jLNgh][kLNgh]);
// 				}
// 				if(extendFace){
// 					if(lsVal>0)positiveNodes.push_back(lsNode(i,j,k,lsVal));
// 					else negativeNodes.push_back(lsNode(i,j,k,lsVal));
// 				}
// 				// right face (rare)
// 				extendFace=false;
// 				if((!(ijk[dim]==IJK[dim]-1)) && extendThis.s[iRNgh][jRNgh][kRNgh]==lsNode::locked){
// 					extendFace=true;
// 					lsVal=0.5*(s[i][j][k]+s[iRNgh][jRNgh][kRNgh]);
// 				}
// 				if(extendFace){
// 					if(lsVal>0)positiveNodes.push_back(lsNode(iRNgh,jRNgh,kRNgh,lsVal));
// 					else negativeNodes.push_back(lsNode(iRNgh,jRNgh,kRNgh,lsVal));
// 				}
// 			}
// 		}
// 		std::sort(positiveNodes.begin(),positiveNodes.end());// sorts in reverse of what we need.
// 		std::sort(negativeNodes.begin(),negativeNodes.end());// sorts in exact order we need.
// 		for(unsigned int n=0;n<negativeNodes.size();n++){//compute the extended negative
// 			int i=negativeNodes[n].i;
// 			int j=negativeNodes[n].j;
// 			int k=negativeNodes[n].k;
// 			REAL iLoc=i+loc[dim][0];
// 			REAL jLoc=j+loc[dim][1];
// 			REAL kLoc=k+loc[dim][2];
// 			// compute the normal so information only flows from higher to lower
// 			// first find the boundary levelset value at the point we are at lsM, then for the
// 			// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 			// or lsL is > lsM then use the maximum to compute the one sided difference.
// 			// If both are smaller, then we set LS_x=0 because no velocity information
// 			// should be flowing into that point from the x direction.
// 			// do similar for the LS_y, and LS_z
// 			REAL lsM=ls.lirp(iLoc,jLoc,kLoc);
// 			REAL lsL=ls.lirp(iLoc-1,jLoc,kLoc);
// 			REAL lsR=ls.lirp(iLoc+1,jLoc,kLoc);
// 			//REAL LS_x=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_x;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_x=0.0;
// 			} else if(lsL>lsR){
// 				LS_x=lsM-lsL;
// 			} else LS_x=lsR-lsM;
			
// 			lsL=ls.lirp(iLoc,jLoc-1,kLoc);
// 			lsR=ls.lirp(iLoc,jLoc+1,kLoc);
// 			//REAL LS_y=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_y;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_y=0.0;
// 			} else if(lsL>lsR){
// 				LS_y=lsM-lsL;
// 			} else LS_y=lsR-lsM;
			
// 			lsL=ls.lirp(iLoc,jLoc,kLoc-1);
// 			lsR=ls.lirp(iLoc,jLoc,kLoc+1);
// 			//REAL LS_z=myfmax(lsL,lsR)<=lsM?0.0:lsL>lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_z;
// 			if(myfmax(lsL,lsR)<=lsM){
// 				LS_z=0.0;
// 			} else if(lsL>lsR){
// 				LS_z=lsM-lsL;
// 			} else LS_z=lsR-lsM;
// 			// Now we solve the equation LS_x*U_x + LS_y*U_y + LS_z*U_z = 0
// 			// we use upwind differencing to solve the velocity derivatives
// 			// and the LS derivatives are what we use to decide the upwind direction.
// 			// So.  If LS_x >0 we use U_x=(U[i][j][k]-U[i-1][j][k])/dx
// 			if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 				REAL Numerator=0;
// 				REAL Denominator=0;
// 				if(LS_x){
// 					Numerator+=LS_x<0?
// 					  VEL[dim]->s[i-1][j][k]*LS_x:
// 					  -VEL[dim]->s[i+1][j][k]*LS_x;
// 					Denominator+=LS_x<0?LS_x:-LS_x;
// 				}
// 				if(LS_y){
// 					Numerator+=LS_y<0?VEL[dim]->s[i][j-1][k]*LS_y:-VEL[dim]->s[i][j+1][k]*LS_y;
// 					Denominator+=LS_y<0?LS_y:-LS_y;
// 				}
// 				if(LS_z){
// 					Numerator+=LS_z<0?VEL[dim]->s[i][j][k-1]*LS_z:-VEL[dim]->s[i][j][k+1]*LS_z;
// 					Denominator+=LS_z<0?LS_z:-LS_z;
// 				}
// 				VEL[dim]->s[i][j][k]=Numerator/Denominator;
// 			} else {
// 				dim?dim==1?numFailedVn++:numFailedWn++:numFailedUn++;
// 			}
// 		}
		
// 		for(int n=positiveNodes.size()-1;n>=0;n--){//compute the extended Positive
// 			int i=positiveNodes[n].i;
// 			int j=positiveNodes[n].j;
// 			int k=positiveNodes[n].k;
// 			REAL iLoc=i+loc[dim][0];
// 			REAL jLoc=j+loc[dim][1];
// 			REAL kLoc=k+loc[dim][2];
// 			// compute the normal so information only flows from low to higher
// 			// first find the levelset value at the point we are at lsM, then for the
// 			// X derivative (LS_x) find right and left values lsR and lsL.  If either lsR
// 			// or lsL is < lsM then use the minimim to compute the one sided difference.
// 			// If both are larger, then we set LS_x=0 because no velocity information
// 			// should be flowing into that point from the x direction.
// 			// do similar for the LS_y, and LS_z
// 			REAL lsM=lirp(iLoc,jLoc,kLoc);
// 			REAL lsL=lirp(iLoc-1,jLoc,kLoc);
// 			REAL lsR=lirp(iLoc+1,jLoc,kLoc);
// 			//REAL LS_x=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_x;
// 			if(myfmin(lsL,lsR)>=lsM){
// 				LS_x=0.0;
// 			} else if(lsL<lsR){
// 				LS_x=lsM-lsL;
// 			} else LS_x=lsR-lsM;
			
// 			lsL=ls.lirp(iLoc,jLoc-1,kLoc);
// 			lsR=ls.lirp(iLoc,jLoc+1,kLoc);
// 			//REAL LS_y=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_y;
// 			if(myfmin(lsL,lsR)>=lsM){
// 				LS_y=0.0;
// 			} else if(lsL<lsR){
// 				LS_y=lsM-lsL;
// 			} else LS_y=lsR-lsM;
			
// 			lsL=ls.lirp(iLoc,jLoc,kLoc-1);
// 			lsR=ls.lirp(iLoc,jLoc,kLoc+1);
// 			//REAL LS_z=myfmin(lsL,lsR)>=lsM?0.0:lsL<lsR?lsM-lsL:lsR-lsM;
// 			REAL LS_z;
// 			if(myfmin(lsL,lsR)>=lsM){
// 				LS_z=0.0;
// 			} else if(lsL<lsR){
// 				LS_z=lsM-lsL;
// 			} else LS_z=lsR-lsM;
// 			// Now we solve the equation LS_x*U_x + LS_y*U_y + LS_z*U_z = 0
// 			// we use upwind differencing to solve the velocity derivatives
// 			// and the LS derivatives are what we use to decide the upwind direction.
// 			// So.  If LS_x >0 we use U_x=(U[i][j][k]-U[i-1][j][k])/dx
// 			if(LS_x||LS_y||LS_z){//ignore divide by zero errors
// 				REAL Numerator=0;
// 				REAL Denominator=0;
// 				if(LS_x){
// 					Numerator+=LS_x>0?VEL[dim]->s[i-1][j][k]*LS_x:-VEL[dim]->s[i+1][j][k]*LS_x;
// 					Denominator+=LS_x>0?LS_x:-LS_x;
// 				}
// 				if(LS_y){
// 					Numerator+=LS_y>0?VEL[dim]->s[i][j-1][k]*LS_y:-VEL[dim]->s[i][j+1][k]*LS_y;
// 					Denominator+=LS_y>0?LS_y:-LS_y;
// 				}
// 				if(LS_z){
// 					Numerator+=LS_z>0?VEL[dim]->s[i][j][k-1]*LS_z:-VEL[dim]->s[i][j][k+1]*LS_z;
// 					Denominator+=LS_z>0?LS_z:-LS_z;
// 				}
// 				VEL[dim]->s[i][j][k]=Numerator/Denominator;
// 			} else {
// 				dim?dim==1?numFailedVp++:numFailedWp++:numFailedUp++;
// 			}
// 		}
// 		nodes.clear();
// 	}
// 	if(numFailedUn||numFailedVn||numFailedWn||numFailedUp||numFailedVp||numFailedWp){
// 		printf("SYS_ExtendStagGrid failed:");
// 		if(numFailedUp)printf(" Up<%i>",numFailedUp);
// 		if(numFailedVp)printf(" Vp<%i>",numFailedVp);
// 		if(numFailedWp)printf(" Wp<%i>",numFailedWp);
// 		if(numFailedUn)printf(" Un<%i>",numFailedUn);
// 		if(numFailedVn)printf(" Vn<%i>",numFailedVn);
// 		if(numFailedWn)printf(" Wn<%i>",numFailedWn);
// 		printf("\n");
// 	}
// }*/

double reinitLevelSetTimer = 0;

void LevelSet::Reinitialize(){
	Timer tmr;
	tmr.tick();
	InitBand();
	//InitializeInterface();
	InitializeInterface(true);
	//RedistancePDE(4,.5);//,true);
	RedistanceBands(true, true, true);
	//RedistanceBands();
	RedistanceCrudelyPastBands();
	//ClampDistancePastBands();
	reinitLevelSetTimer = tmr./*Elapsed*/tick();
}

// void LevelSet::SetNarrowBandWidth(float CFL){
// 	narrowBandWidth=(int)myfmax((int)(CFL+1.0),minNarrowBandWidth);
// }

// int LevelSet::SaveSurface(char *fname, bool save_bands_only){
// 	// This saves out a band, the header is simple, look at the printf.
// 	// RETURN 0 if it was able to find the file 
// 	// or return 1 if it could not find the file.
//   FILE* fp=fopen(fname,"wb");
// 	if(!fp)return(-0x01);
//   int _sizeof=sizeof(s[0][0][0]);
//   fprintf(fp,"v1.0\nbytes: %i\n%i %i %i\nband\n",_sizeof,I,J,K);
// 	fflush(fp);
// 	int error=0;
// 	std::vector<short> i;
// 	std::vector<short> nj;
// 	std::vector<short> j;
// 	std::vector<short> nk;
// 	std::vector<short> k;
// 	std::vector<float> data;
// 	field3d<char> SaveThis(I,J,K,0);
// 	int cur_i=-1;
// 	int buffer_size=2;
// 	// the surface band has been reinitilaized and maybe corrected, so it my not
// 	// hold the only cells with neighbors with opposite neighbors.
// 	// but this function should not change the state of the level set object.
// 	if(save_bands_only){
// 		for(int m=0;m<3;m++){
// 			std::vector<lsNode>& curBand=*(allBands[m]);
// 			for(unsigned int n=0;n<curBand.size();n++){
// 				int ii=curBand[n].i;
// 				int jj=curBand[n].j;
// 				int kk=curBand[n].k;
// 				SaveThis.s[ii][jj][kk]=1;
// 			}
// 		}
// 	} else for(int ii=0;ii<I;ii++)for(int jj=0;jj<J;jj++)for(int kk=0;kk<K;kk++){
// 		bool hasOpositeNeighbor=false;
// 		if(s[ii][jj][kk]>0){
//       if(ii!=0&&s[ii-1][jj][kk]<=0)hasOpositeNeighbor=true;
//       else if(jj!=0&&s[ii][jj-1][kk]<=0)hasOpositeNeighbor=true;
//       else if(kk!=0&&s[ii][jj][kk-1]<=0)hasOpositeNeighbor=true;
//       else if(ii!=(I-1)&&s[ii+1][jj][kk]<=0)hasOpositeNeighbor=true;
//       else if(jj!=(J-1)&&s[ii][jj+1][kk]<=0)hasOpositeNeighbor=true;
//       else if(kk!=(K-1)&&s[ii][jj][kk+1]<=0)hasOpositeNeighbor=true;
// 		} else {// fluid.s[i][j][k]<=0
//       if(ii!=0&&s[ii-1][jj][kk]>0)hasOpositeNeighbor=true;
//       else if(jj!=0&&s[ii][jj-1][kk]>0)hasOpositeNeighbor=true;
//       else if(kk!=0&&s[ii][jj][kk-1]>0)hasOpositeNeighbor=true;
//       else if(ii!=(I-1)&&s[ii+1][jj][kk]>0)hasOpositeNeighbor=true;
//       else if(jj!=(J-1)&&s[ii][jj+1][kk]>0)hasOpositeNeighbor=true;
//       else if(kk!=(K-1)&&s[ii][jj][kk+1]>0)hasOpositeNeighbor=true;
// 		}
// 		if(hasOpositeNeighbor){
// 			int iil=ii-buffer_size;
// 			if(iil<0)iil=0;
// 			int iir=ii+buffer_size;
// 			if(iir>I-1)iir=I-1;
// 			int jjl=jj-buffer_size;
// 			if(jjl<0)jjl=0;
// 			int jjr=jj+buffer_size;
// 			if(jjr>J-1)jjr=J-1;
// 			int kkl=kk-buffer_size;
// 			if(kkl<0)kkl=0;
// 			int kkr=kk+buffer_size;
// 			if(kkr>K-1)kkr=K-1;
			
// 			for(int iii=iil;iii<=iir;iii++){
// 				for(int jjj=jjl;jjj<=jjr;jjj++){
// 					for(int kkk=kkl;kkk<=kkr;kkk++){
// 						SaveThis.s[iii][jjj][kkk]=1;
// 					}
// 				}
// 			}
// 		}
// 	}
	
// 	for(int ii=0;ii<I;ii++)for(int jj=0;jj<J;jj++)for(int kk=0;kk<K;kk++){
// 		if(SaveThis.s[ii][jj][kk]){
// 			if(cur_i!=ii){
// 				// new i
// 				cur_i=ii;
// 				i.push_back(cur_i);
// 				nj.push_back(1);
// 				j.push_back(jj);
// 				nk.push_back(1);
// 				k.push_back(kk);
// 				data.push_back(s[ii][jj][kk]);
// 			} else if(j.back()!=jj){
// 				// new j
// 				nj.back()++;
// 				j.push_back(jj);
// 				nk.push_back(1);
// 				k.push_back(kk);
// 				data.push_back(s[ii][jj][kk]);
// 			} else if(k.back()!=kk){
// 				// new k
// 				nk.back()++;
// 				k.push_back(kk);
// 				data.push_back(s[ii][jj][kk]);
// 			}
// 		}
// 	}
// 	unsigned long cur_j=0;
// 	unsigned long cur_k=0;
//   for(unsigned int ii=0;ii<i.size();ii++){
//     error+=(1-fwrite(&(i[ii]),sizeof(short),1,fp));
//     error+=(1-fwrite(&(nj[ii]),sizeof(short),1,fp));
// 		for(int jj=0;jj<nj[ii];jj++){
// 	    error+=(1-fwrite(&(j[cur_j]),sizeof(short),1,fp));
// 	    error+=(1-fwrite(&(nk[cur_j]),sizeof(short),1,fp));
// 			for(int kk=0;kk<nk[cur_j];kk++){
// 	    	error+=(1-fwrite(&(k[cur_k]),sizeof(short),1,fp));
// 	    	error+=(1-fwrite(&(data[cur_k]),sizeof(float),1,fp));
// 				cur_k++;
// 			}
// 			cur_j++;
// 		}
// 		fflush(fp);
//   }
//   fclose(fp);
// 	fflush(0);
// 	return(error);
// }
// int LevelSet::LoadSurface(char *fname, field3d<char>* mask_p){
//   FILE* fp=fopen(fname,"rb");
// 	if(!fp)return(-0x01);
//   int _sizeof;
// 	char* aLine=0;
// 	size_t len=0;
//   fscanf(fp,"v1.0\nbytes: %i\n",&_sizeof);
// 	//ssize_t read=
// 	getline(&aLine,&len,fp);
// 	int inI,inJ,inK;
// 	sscanf(aLine,"%i %i %i",&inI,&inJ,&inK);
// 	I=J=K=0;//so resize won't copy
// 	resize(inI,inJ,inK);
// 	getline(&aLine,&len,fp);// get the "band"
// 	if(aLine)free(aLine);
// 	if(mask_p){ 
// 		int maskI=std::max(mask_p->I,I);
// 		int maskJ=std::max(mask_p->J,J);
// 		int maskK=std::max(mask_p->K,K);
// 		mask_p->resize(maskI,maskJ,maskK);
// 		(*mask_p)=0;
// 	}
	 
// 	allBands[0]=&bandSurface;
// 	bandSurface.clear();

// 	// Here is where we load the Surface level set band
// 	// if a node is positive and has a negative node next to it then we add
// 	// it, and visa versa for negative nodes.
// 	short i,j,k,nj,nk;
// 	float data;
// 	while(fread(&i,sizeof(short),1,fp)){// while not end of file, read in the current i value
// 		fread(&nj,sizeof(short),1,fp);
// 		for(int jj=0;jj<nj;jj++){
// 			fread(&j,sizeof(short),1,fp);
// 			fread(&nk,sizeof(short),1,fp);
// 			for(int kk=0;kk<nk;kk++){
// 				fread(&k,sizeof(short),1,fp);
// 				fread(&data,sizeof(float),1,fp);
// 				s[i][j][k]=data;
//       	bandSurface.push_back(lsNode(i,j,k,s[i][j][k]));
// 				if(mask_p)mask_p->s[i][j][k]=1;
// 			}
// 		}
//   }
// 	fclose(fp);
	

// 	InitBand(3,true);
//   RedistanceBands(true,false);
//   RedistanceBands(false,true);
// //  RedistanceBands(true,false);
// 	RedistanceCrudelyPastBands();
	
// 	return(0);
// }


// int LevelSet::LoadUnknown(char *fname, field3d<char>* mask_p){
//   FILE* fp=fopen(fname,"rb");
// 	if(!fp)return(-0x01);
// 	int retVal;
// 	char* aLine=0;
// 	size_t len=0;
// 	// There are three types of files that this funciton knows how to load.  field3d, mask3d, and band3d.
// 	// the header of mask3d files is written with
// 	//  fprintf(fp,"v1.0\nbytes: %i\n%i %i %i\nmask\n",_sizeof,I,J,K);
// 	// the header of the band3d files is written with
// 	//  fprintf(fp,"v1.0\nbytes: %i\n%i %i %i\nband\n",_sizeof,I,J,K);
// 	// for field field3d files it is written with
// 	//  fprintf(fp,"v1.0\ntype: %s\nbytes: %i\n%i %i %i\n",typeid(T).name(),_sizeof,I,J,K);
// 	getline(&aLine,&len,fp); // get the version number line.
// 	getline(&aLine,&len,fp); // bytes: or type: will determin if it is a mask or a field
// 	if(!strncmp("type:",aLine,5)){// field3d
// 		retVal=Load(fname);
// 		if(mask_p){ 
// 			int maskI=std::max(mask_p->I,I);
// 			int maskJ=std::max(mask_p->J,J);
// 			int maskK=std::max(mask_p->K,K);
// 			mask_p->resize(maskI,maskJ,maskK);
// 			(*mask_p)=1;
// 		}
// 		InitBand(3);
// 	} else if(!strncmp("bytes:",aLine,6)){
// 		getline(&aLine,&len,fp);// get the I J K
// 		getline(&aLine,&len,fp);// get mask or band
// 		if(!strncmp("mask",aLine,4)){
// 			retVal=LoadSurface(fname,mask_p);
// 		} else if(!strncmp("band",aLine,4)){
// 			retVal=LoadSurface(fname,mask_p);
// 		} else {
// 			retVal=-0x02;
// 		}
// 	} else {
// 		retVal=-0x03;
// 	}
// 	fclose(fp);
// 	if(aLine)free(aLine);
// 	return retVal;
// }

// /* Private Members */

// /* LevelSet::AdvanceBFECC--Back and Forth Error Compensation and Correction
// //                         Advection for all level set values in the narrow band.
// //
// */
// void LevelSet::AdvanceBFECC(float dt, float dxInv, StaggeredGrid& velocity){
// 	// Set the Scale: - because we march backwards along the characteristics,
// 	// dt becuase that is the time we have to look, dx_inv because the 
// 	// computational cell width is dx while the MAC cell width is 1
	
// 	// changed by VIVEK, to avoid allocation every time step
// 	//ScalarGrid LF(*this);
// 	static ScalarGrid LF;
// 	LF = *this;
	
// 	float scale=-dt*dxInv;
// 	vec scaledVelocity;
// 	// do one fordard, save it in current
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity=scale*(velocity(i,j,k));
// 			s[i][j][k]=LF.lirp(i+scaledVelocity.x,j+scaledVelocity.y,k+scaledVelocity.z);
// 		}
// 	}
// 	// do a Backwards on s, and save phi^n + 0.5*[phi^n-LF] in LF
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity=-scale*(velocity(i,j,k));
// 			LF.s[i][j][k] += 0.5*(LF.s[i][j][k] -lirp(i+scaledVelocity.x,j+scaledVelocity.y,k+scaledVelocity.z));
// 		}
// 	}
// 	// run a forward on LF and save it in s
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity=scale*(velocity(i,j,k));
// 			s[i][j][k]=LF.lirp(i+scaledVelocity.x,j+scaledVelocity.y,k+scaledVelocity.z);
// 		}
// 	}
	
// }
// /* LevelSet::AdvanceSL--Semi-Lagrangian Advection for all level set values in the narrow band.
// //
// */
// void LevelSet::AdvanceSL(float dt, float dxInv, StaggeredGrid& velocity){
// 	// Set the Scale: - because we march backwards along the characteristics,
// 	// dt becuase that is the time we have to look, dx_inv because the 
// 	// computational cell width is dx while the MAC cell width is 1
// 	float scale=-dt*dxInv;
// 	vec scaledVelocity;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity=scale*(velocity(i,j,k));
// 			curBand[n].ls=lirp(i+scaledVelocity.x,j+scaledVelocity.y,k+scaledVelocity.z);
// 		}
// 	}
// 	// now that all the values have been computed, we need to assign them
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			s[curBand[n].i][curBand[n].j][curBand[n].k]=curBand[n].ls;
// 		}
// 	}
// }

// void LevelSet::AdvanceScalarBandsBFECC(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids){
// 	// Set the Scale: - because we march backwards along the characteristics,
// 	// dt becuase that is the time we have to look, dx_inv because the 
// 	// computational cell width is dx while the MAC cell width is 1
	
// 	unsigned int bandSize=bandSurface.size()+bandPositive.size()+bandNegative.size();
// 	std::vector<float*> saved;
// 	std::vector<float*> LF;
// 	std::vector<float> rng_min;
// 	std::vector<float> rng_max;
// 	for(unsigned int n=0;n<grids.size();n++){
// 		saved.push_back(new float[bandSize]);
// 		LF.push_back(new float[bandSize]);
// 		rng_min.push_back(1e+36);
// 		rng_max.push_back(-1e+36);
// 	}
// 	float scale=-dt*dxInv;
// 	vec* scaledVelocity=new vec[bandSize];
// 	// Save the current values of the grids, and Calulate the first foreard L
// 	// ALSO ! BFECC can make something in a range (say [0 1]) go outside that range
// 	//  so find the range and clamp any results to inside it at the end.
// 	int counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity[counter]=scale*(velocity(i,j,k));
// 			for(unsigned int g=0;g<LF.size();g++){
// 				saved[g][counter]=grids[g]->s[i][j][k];
// 				rng_min[g]=myfmin(rng_min[g],saved[g][counter]);
// 				rng_max[g]=myfmax(rng_max[g],saved[g][counter]);
// 				LF[g][counter]=grids[g]->lirp(i+scaledVelocity[counter].x,j+scaledVelocity[counter].y,k+scaledVelocity[counter].z);
// 			}
// 			counter++;
// 		}
// 	}
// 	// Store the forward LF in the grid
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<LF.size();g++)grids[g]->s[i][j][k]=LF[g][counter];
// 			counter++;
// 		}
// 	}
// 	// Solve phi^n+0.5*(phi^n-L(-u,LF)) and store in LF
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<LF.size();g++){
// 				LF[g][counter]=saved[g][counter]+0.5*(saved[g][counter]-grids[g]->lirp(i-scaledVelocity[counter].x,j-scaledVelocity[counter].y,k-scaledVelocity[counter].z));
// 			}
// 			counter++;
// 		}
// 	}
// 	// Store LF in the grid
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<LF.size();g++)grids[g]->s[i][j][k]=LF[g][counter];
// 			counter++;
// 		}
// 	}
// 	// Solve LF again on the grid
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<LF.size();g++){
// 				LF[g][counter]=grids[g]->lirp(i+scaledVelocity[counter].x,j+scaledVelocity[counter].y,k+scaledVelocity[counter].z);
// 			}
// 			counter++;
// 		}
// 	}
// 	// Store LF in the grid
// 	// make sure to clam the values.
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<saved.size();g++){
// 				float clamped=LF[g][counter];
// 				if(clamped < rng_min[g])clamped=rng_min[g];
// 				if(clamped > rng_max[g])clamped=rng_max[g];
// 				grids[g]->s[i][j][k]=clamped;
// 			}
// 			counter++;
// 		}
// 	}
	
// 	for(unsigned int g=0;g<LF.size();g++){
// 		delete[](saved[g]);
// 		delete[](LF[g]);
// 	}
// 	delete[](scaledVelocity);
// }
// void LevelSet::AdvanceScalarBandsSL(float dt, float dxInv, StaggeredGrid& velocity, std::vector<ScalarGrid*>& grids){
// 	// Set the Scale: - because we march backwards along the characteristics,
// 	// dt becuase that is the time we have to look, dx_inv because the 
// 	// computational cell width is dx while the MAC cell width is 1
	
// 	unsigned int bandSize=bandSurface.size()+bandPositive.size()+bandNegative.size();
// 	std::vector<float*> saved;
// 	for(unsigned int n=0;n<grids.size();n++)saved.push_back(new float[bandSize]);
// 	float scale=-dt*dxInv;
// 	vec scaledVelocity;
// 	// Save the current values of the grids
// 	int counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			scaledVelocity=scale*(velocity(i,j,k));
// 			for(unsigned int g=0;g<saved.size();g++)saved[g][counter]=grids[g]->lirp(i+scaledVelocity.x,j+scaledVelocity.y,k+scaledVelocity.z);
// 			counter++;
// 		}
// 	}
// 	// now that all the values have been computed, we need to assign them
// 	counter=0;
// 	for(int m=0;m<3;m++){
// 		std::vector<lsNode>& curBand=*(allBands[m]);
// 		for(unsigned int n=0;n<curBand.size();n++){
// 			int i=curBand[n].i;
// 			int j=curBand[n].j;
// 			int k=curBand[n].k;
// 			for(unsigned int g=0;g<saved.size();g++)grids[g]->s[i][j][k]=saved[g][counter];
// 			counter++;
// 		}
// 	}
// 	for(unsigned int g=0;g<saved.size();g++)delete[](saved[g]);
// }

/* LevelSet::InitializeInterface--make the surface of the level set a 
// SDF.  This function follows the improved
// treatment of the initial band found in "Spatially Adaptive Techniques
// for level set methods and Incompressible Flow," Losasso, et.al 2004
//
// Before you call this function it is assumed that you have called
// InitBand, to find the cells that we want
// to run this on.
*/
void LevelSet::InitializeInterface(bool conditionalredistance, int maxIter, float tolerance){
	// First thing to do is get acceptable values for the crossing nodes.
	// This is done in a few steps:
	// Step 1: find up to three corssing points on the cartesion axis, if there
	//  more than one on an axis take the smallest, set the smallest of the three
	//  to the threshold for the next step
	// Step 2: Itterativly move a point along the local normal from the starting
	//  point to the surface, if it converges use the dist to the point as new val
	// Step 3: If step 2 does not converge, for the dist is > the threshold, then
	//  use the up to three points from above to get the values.  1 value and use
	//  dist to that point.  2 values and use distance to the line.  3 values
	//  and use the distance to the plane.
	REAL tempDist;
	int crossingsZero=0;// Keep track of usually fetal error, print how many if there were any
	// if there are any zero corssings, that usually means corrup data.
	for(unsigned int n=0;n<bandSurface.size();n++){
		int numAxisPoints=0;
		// Step 1: find the smallest crossing values
		vec crossingDists(-2,-2,-2);// -2 if no crossing, [-1,0] val if crossing to left, below, behind, [0,1] otherwise
		int i=bandSurface[n].i;
		int j=bandSurface[n].j;
		int k=bandSurface[n].k;
		if(at(i,j,k)>0){// at a positive node
			// x-cartesion axis
      if(i!=0&&at(i-1,j,k)<=0){// negative neighbor to left
				numAxisPoints++;
				crossingDists.x=at(i,j,k)/(at(i-1,j,k)-at(i,j,k));
			}
			if(i!=(I-1)&&at(i+1,j,k)<=0){// negative neighbor to right
				tempDist=at(i,j,k)/(at(i,j,k)-at(i+1,j,k));
				if((-crossingDists.x)>tempDist){
					if(crossingDists.x==-2)numAxisPoints++;
					crossingDists.x=tempDist;
				}
			}
			// y-cartesion axis
      if(j!=0&&at(i,j-1,k)<=0){// negative neighbor below
				numAxisPoints++;
				crossingDists.y=at(i,j,k)/(at(i,j-1,k)-at(i,j,k));
			}
			if(j!=(J-1)&&at(i,j+1,k)<=0){// negative neighbor above
				tempDist=at(i,j,k)/(at(i,j,k)-at(i,j+1,k));
				if((-crossingDists.y)>tempDist){
					if(crossingDists.y==-2)numAxisPoints++;
					crossingDists.y=tempDist;
				}
			}
			// z-cartesion axis
      if(k!=0&&at(i,j,k-1)<=0){// negative neighbor behind
				numAxisPoints++;
				crossingDists.z=at(i,j,k)/(at(i,j,k-1)-at(i,j,k));
			}
			if(k!=(K-1)&&at(i,j,k+1)<=0){// negative neighbor in front
				tempDist=at(i,j,k)/(at(i,j,k)-at(i,j,k+1));
				if((-crossingDists.z)>tempDist){
					if(crossingDists.z==-2)numAxisPoints++;
					crossingDists.z=tempDist;
				}
			}
		} else {// fluid.s[i][j][k])<=0, at a negative node
			// x-cartesion axis
      if(i!=0&&at(i-1,j,k)>0){// positive neighbor to left
				numAxisPoints++;
				crossingDists.x=at(i,j,k)/(at(i-1,j,k)-at(i,j,k));
			}
			if(i!=(I-1)&&at(i+1,j,k)>0){// positive neighbor to right
				tempDist=at(i,j,k)/(at(i,j,k)-at(i+1,j,k));
				if((-crossingDists.x)>tempDist){
					if(crossingDists.x==-2)numAxisPoints++;
					crossingDists.x=tempDist;
				}
			}
			// y-cartesion axis
      if(j!=0&&at(i,j-1,k)>0){// positive neighbor below
				numAxisPoints++;
				crossingDists.y=at(i,j,k)/(at(i,j-1,k)-at(i,j,k));
			}
			if(j!=(J-1)&&at(i,j+1,k)>0){// positive neighbor above
				tempDist=at(i,j,k)/(at(i,j,k)-at(i,j+1,k));
				if((-crossingDists.y)>tempDist){
					if(crossingDists.y==-2)numAxisPoints++;
					crossingDists.y=tempDist;
				}
			}
			// z-cartesion axis
      if(k!=0&&at(i,j,k-1)>0){// positive neighbor behind
				numAxisPoints++;
				crossingDists.z=at(i,j,k)/(at(i,j,k-1)-at(i,j,k));
			}
			if(k!=(K-1)&&at(i,j,k+1)>0){// positive neighbor in front
				tempDist=at(i,j,k)/(at(i,j,k)-at(i,j,k+1));
				if((-crossingDists.z)>tempDist){
					if(crossingDists.z==-2)numAxisPoints++;
					crossingDists.z=tempDist;
				}
			}
		}
		
		// Step 2: move a point out to find the level set crossing.
		REAL threshold=myfmin(fabs(crossingDists.x),fabs(crossingDists.y));
		threshold=myfmin(threshold,fabs(crossingDists.z));
		vec P(i,j,k);
		vec norm=N(i,j,k);
		REAL dist_to_ls=at(i,j,k);
		int iter;
		for(iter=0;iter<maxIter;iter++){
			P-=norm*dist_to_ls;
			dist_to_ls=lirp(P);
			if(fabs(dist_to_ls)<tolerance){
				break;
			}
			norm=lirpN(P);
		}
		P.x-=i;
		P.y-=j;
		P.z-=k;
		REAL dist_to_node=P.Length();
		if(iter==maxIter||dist_to_node>threshold){
			// Step 3 is here, becuase step 2 failed
			if(numAxisPoints==1){
				// use the linear interpolated distance along the axis, we already
				// have this distance stored in threshold.
				dist_to_node=threshold;
			}else if(numAxisPoints==2){
				// Use the distance to the line defined by the two intersection points
				// as the distance to the level set.  This is a very simplified computation
				// of distance from a poin to a line.  We know the distances d1 and d2
				// are  on different axis, and with symitries we don't need to worry
				// about what plane they lay in... if x1 and x2 are the points on the
				// line, and x0 is the point we want the distance to, then we could use
				// the formula:
				// d = |(x2-x1)^(x1-x0)|/|X2-X1|
				// but we set x0=0,x1=(d1,0,0),x2=(0,d2,0) we can get...
				REAL d1=crossingDists.x;
				REAL d2=crossingDists.y;
				if(d1==-2)d1=crossingDists.z;
				else if(d2==-2)d2=crossingDists.z;
				d1*=d1;
				d2*=d2;
				dist_to_node=((d1+d2)?sqrt(d1*d2/(d1+d2)):0);
			}else if(numAxisPoints==3){
				// you could use SqrDist_ToTriangle, but we know a lot about the 3
				// points that we need the distance to, so don't use overkill.
				norm.x=crossingDists.y*crossingDists.z;
				norm.y=crossingDists.x*crossingDists.z;
				norm.z=crossingDists.x*crossingDists.y;
				REAL Nlen=norm.Length();
				dist_to_node=Nlen?fabs(crossingDists.x*norm.x)/Nlen:0;
			} else {
				dist_to_node=fabs(at(i,j,k));
				crossingsZero++;//printf("         crossings=%i\n",numAxisPoints);
			}
		}
		bandSurface[n].ls=(at(i,j,k)>0)?dist_to_node:-dist_to_node;
	}
	if(crossingsZero){printf("Num Zero Crossings = %i, BIG ERROR!\n",crossingsZero);fflush(0);}
	// Now assign the computed values.
	if(!conditionalredistance)for(unsigned int n=0;n<bandSurface.size();n++){
			at(bandSurface[n].i,bandSurface[n].j,bandSurface[n].k)=bandSurface[n].ls;
	} else for(unsigned int n=0;n<bandSurface.size();n++){
		//dupont, liu, bfecc for sl level set method, 05, p.9
		int i=bandSurface[n].i;
		int j=bandSurface[n].j;
		int k=bandSurface[n].k;
		bool assignnew=false;
		if(i && fabs(at(i,j,k)-at(i-1,j,k))>1.1) assignnew=true;
		else if(i<I-1 && fabs(at(i,j,k)-at(i+1,j,k))>1.1) assignnew=true;
		else if(j && fabs(at(i,j,k)-at(i,j-1,k))>1.1) assignnew=true;
		else if(j<J-1 && fabs(at(i,j,k)-at(i,j+1,k))>1.1) assignnew=true;
		else if(k && fabs(at(i,j,k)-at(i,j,k-1))>1.1) assignnew=true;
		else if(k<K-1 && fabs(at(i,j,k)-at(i,j,k+1))>1.1) assignnew=true;
		
		if(assignnew)at(i,j,k)=bandSurface[n].ls;
	}
}

/* LevelSet::InitBand()--Initialize level set's narrow band
//
// This function first finds all the bandSurface nodes, those nodes with
// at least one neighbor node of differing sign.  It then finds a band
// of size BandWidth the the negative and positive side and puts them
// in bandNegative and bandPositive.
*/
void LevelSet::InitBand(int BandWidth, bool haveSurface){
  
	allBands[0]=&bandSurface;
	allBands[1]=&bandPositive;
	allBands[2]=&bandNegative;
	// changed by VIVEK, to avoid allocation every time step
	//if(!haveSurface)bandSurface.clear();
	//bandNegative.clear();
	//bandPositive.clear();
	if(!haveSurface)bandSurface.resize(0);
	bandNegative.resize(0);
	bandPositive.resize(0);

	
	// The Visited field will keep track of what nodes are already in the band
	// so that we don't add them more than once.  It starts out with a value 
	// higher than the highest band, and is marked with the band level of the
	// node that is inserted
	// changed by VIVEK, to avoid allocation every time step
	//field3d<int> Visited(I,J,K,BandWidth+1);
	static field3d<int> Visited;
	Visited.resize(I,J,K);
	Visited = BandWidth;

	// Here is where we find the Surface level set band
	// if a node is positive and has a negative node next to it then we add
	// it, and visa versa for negative nodes.
	if(haveSurface){
		for(unsigned int b=0;b<bandSurface.size();b++)Visited.at(bandSurface[b].i,bandSurface[b].j,bandSurface[b].k)=0;
	} else for(int i=0;i<I;i++)for(int j=0;j<J;j++)for(int k=0;k<K;k++){
		bool hasOpositeNeighbor=false;
		if(at(i,j,k)>0){
      if(i!=0&&at(i-1,j,k)<=0)hasOpositeNeighbor=true;
      else if(j!=0&&at(i,j-1,k)<=0)hasOpositeNeighbor=true;
      else if(k!=0&&at(i,j,k-1)<=0)hasOpositeNeighbor=true;
      else if(i!=(I-1)&&at(i+1,j,k)<=0)hasOpositeNeighbor=true;
      else if(j!=(J-1)&&at(i,j+1,k)<=0)hasOpositeNeighbor=true;
      else if(k!=(K-1)&&at(i,j,k+1)<=0)hasOpositeNeighbor=true;
		} else {// fluid.s[i][j][k]<=0
      if(i!=0&&at(i-1,j,k)>0)hasOpositeNeighbor=true;
      else if(j!=0&&at(i,j-1,k)>0)hasOpositeNeighbor=true;
      else if(k!=0&&at(i,j,k-1)>0)hasOpositeNeighbor=true;
      else if(i!=(I-1)&&at(i+1,j,k)>0)hasOpositeNeighbor=true;
      else if(j!=(J-1)&&at(i,j+1,k)>0)hasOpositeNeighbor=true;
      else if(k!=(K-1)&&at(i,j,k+1)>0)hasOpositeNeighbor=true;
		}
		if(hasOpositeNeighbor){
			Visited.at(i,j,k)=0;
      bandSurface.push_back(lsNode(i,j,k,at(i,j,k)));
    }
  }

	// ok, bandSurface has all the nodes that it needs. Now we need to find the
	// other bands.  Find the first band by looking off of bandSurface.
	int searchBand=1; // This is the current band we are searching for.
	int ii,jj,kk;
	int n,i,j,k;
	float addme=.5;
	if(BandWidth>0)for(n=0;n<(int)bandSurface.size();n++){
		i=bandSurface[n].i;
		j=bandSurface[n].j;
		k=bandSurface[n].k;
		ii=i-1;
		jj=j;
		kk=k;
		if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
		ii=i;
		jj=j-1;
		kk=k;
		if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
		ii=i;
		jj=j;
		kk=k-1;
		if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
		ii=i+1;
		jj=j;
		kk=k;
		if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
		ii=i;
		jj=j+1;
		kk=k;
		if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
		ii=i;
		jj=j;
		kk=k+1;
		if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
			//if(s[ii][jj][kk]>0) bandPositive.push_back(lsNode(ii,jj,kk,s[ii][jj][kk]));
			if(at(i,j,k)>0) bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			else bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
		}
	}
	// ok, now we have the surface band, and the start of the other two.
	// we will be inserting into the bands, so we need to keep track of the
	// spot to stop searching... in each itteration we only need to search
	// the ones we just found for new ones.
	int beginPositive=0;
	int endPositive=bandPositive.size();
	int beginNegative=0;
	int endNegative=bandNegative.size();
	for(searchBand=2;searchBand<=BandWidth;searchBand++){
		// Search toward the positive side.
		for(n=beginPositive;n<endPositive;n++){
			i=bandPositive[n].i;
			j=bandPositive[n].j;
			k=bandPositive[n].k;
			ii=i-1;
			jj=j;
			kk=k;
			if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j-1;
			kk=k;
			if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j;
			kk=k-1;
			if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i+1;
			jj=j;
			kk=k;
			if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j+1;
			kk=k;
			if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j;
			kk=k+1;
			if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandPositive.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
		}
		beginPositive=endPositive;
		endPositive=bandPositive.size();
		// Search toward the Fluid side.
		for(n=beginNegative;n<endNegative;n++){
			i=bandNegative[n].i;
			j=bandNegative[n].j;
			k=bandNegative[n].k;
			ii=i-1;
			jj=j;
			kk=k;
			if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				// we should never actually find the other sign node by this point
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j-1;
			kk=k;
			if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j;
			kk=k-1;
			if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i+1;
			jj=j;
			kk=k;
			if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j+1;
			kk=k;
			if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
			ii=i;
			jj=j;
			kk=k+1;
			if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
				Visited.at(ii,jj,kk)=searchBand;
			if(haveSurface)at(ii,jj,kk)=at(i,j,k)+(at(i,j,k)>0?addme:-addme);
				bandNegative.push_back(lsNode(ii,jj,kk,at(ii,jj,kk)));
			}
		}
		beginNegative=endNegative;
		endNegative=bandNegative.size();		
		if(beginNegative==endNegative && beginPositive==endPositive)break;// no reason to keep going if we can't
	}
}


/* LevelSet::RedistanceBands--make the level set a signed
// distance function.  This function follows the FMM method with improved
// treatment of the initial band found in "Spatially Adaptive Techniques
// for level set methods and Incompressible Flow," Losasso, et.al 2004
//
// Before you call this function it is assumed that you have called
// InitBand, to find the cells that we want
// to run this on. and some semplance of InitializeInterface to get
// a good initial interface value.
//
// To redistance the entire level set call:
//  InitBand(I+J+K);
//  InitializeInterface();// if you need to
//  RedistanceBands();
//
// To redistance the entire level set SAFELY (if there may be holes in it) call:
//  InitBand(I+J+K);
//  InitializeInterface();// if you need to...
//  RedistanceBands(true,false);
//  RedistanceBands(false,true);
*/
void LevelSet::RedistanceBands(bool doPositive, bool doNegative, bool conditionalredistance){
	// NOW COMES THE ACTUAL FAST MARCHING METHOD PART...
	// Don't forget to assign the new bandSurface values to fluid.s[][][]
	// Initialize a field to keep track of the indices into the heap.
	// We need this because we sometimes need to update tentative values that are
	// already in the heap.	
	//field3d<int> heapIndex(I,J,K,lsNode::out);
	static field3d<int> heapIndex;	// changed by VIVEK, to avoid allocation every time step
	heapIndex.resize(I,J,K);
	heapIndex = lsNode::out;

	// the surface values are computed so they are locked
	for(unsigned int n=0;n<bandSurface.size();n++){
		heapIndex.at(bandSurface[n].i,bandSurface[n].j,bandSurface[n].k)=lsNode::locked;
	}
	if(!conditionalredistance){
		// The Negative and Positive bands need values.
		if(doNegative)for(unsigned int n=0;n<bandNegative.size();n++){
				heapIndex.at(bandNegative[n].i,bandNegative[n].j,bandNegative[n].k)=lsNode::extend;
		}
		if(doPositive)for(unsigned int n=0;n<bandPositive.size();n++){
				heapIndex.at(bandPositive[n].i,bandPositive[n].j,bandPositive[n].k)=lsNode::extend;
		}
	} else {// if(conditionalredistance)
		// The Negative and Positive bands need values.
		if(doNegative)for(unsigned int n=0;n<bandNegative.size();n++){
			int i=bandNegative[n].i;
			int j=bandNegative[n].j;
			int k=bandNegative[n].k;
			bool assignnew=false;
			if(i && fabs(at(i,j,k)-at(i-1,j,k))>1.1) assignnew=true;
			else if(i<I-1 && fabs(at(i,j,k)-at(i+1,j,k))>1.1) assignnew=true;
			else if(j && fabs(at(i,j,k)-at(i,j-1,k))>1.1) assignnew=true;
			else if(j<J-1 && fabs(at(i,j,k)-at(i,j+1,k))>1.1) assignnew=true;
			else if(k && fabs(at(i,j,k)-at(i,j,k-1))>1.1) assignnew=true;
			else if(k<K-1 && fabs(at(i,j,k)-at(i,j,k+1))>1.1) assignnew=true;
			
			float spi=(i<I-1)?at(i+1,j,k):at(i,j,k);
			float spj=(j<J-1)?at(i,j+1,k):at(i,j,k);
			float spk=(k<K-1)?at(i,j,k+1):at(i,j,k);
			float smi=(i)?at(i-1,j,k):at(i,j,k);
			float smj=(j)?at(i,j-1,k):at(i,j,k);
			float smk=(k)?at(i,j,k-1):at(i,j,k);
			float spipj=(i<I-1 && j<J-1)?at(i+1,j+1,k):at(i,j,k);
			float spimj=(i<I-1 && j)?at(i+1,j-1,k):at(i,j,k);
			float smipj=(i && j<J-1)?at(i-1,j+1,k):at(i,j,k);
			float smimj=(i && j)?at(i-1,j-1,k):at(i,j,k);
			float spipk=(i<I-1 && k<K-1)?at(i+1,j,k+1):at(i,j,k);
			float spimk=(i<I-1 && k)?at(i+1,j,k-1):at(i,j,k);
			float smipk=(i && k<K-1)?at(i-1,j,k+1):at(i,j,k);
			float smimk=(i && k)?at(i-1,j,k-1):at(i,j,k);
			float spjpk=(j<J-1 && k<K-1)?at(i,j+1,k+1):at(i,j,k);
			float spjmk=(j<J-1 && k)?at(i,j+1,k-1):at(i,j,k);
			float smjpk=(j && k<K-1)?at(i,j-1,k+1):at(i,j,k);
			float smjmk=(j && k)?at(i,j-1,k-1):at(i,j,k);
			float spipjpk=(i<I-1 && j<J-1 && k<K-1)?at(i+1,j+1,k+1):at(i,j,k);
			float spipjmk=(i<I-1 && j<J-1 && k)?at(i+1,j+1,k-1):at(i,j,k);
			float spimjpk=(i<I-1 && j && k<K-1)?at(i+1,j-1,k+1):at(i,j,k);
			float spimjmk=(i<I-1 && j && k)?at(i+1,j-1,k-1):at(i,j,k);
			float smipjpk=(i && j<J-1 && k<K-1)?at(i-1,j+1,k+1):at(i,j,k);
			float smipjmk=(i && j<J-1 && k)?at(i-1,j+1,k-1):at(i,j,k);
			float smimjpk=(i && j && k<K-1)?at(i-1,j-1,k+1):at(i,j,k);
			float smimjmk=(i && j && k)?at(i-1,j-1,k-1):at(i,j,k);
			
			if((at(i,j,k)<=0&&spi<=0&&spj<=0&&spk<=0&&smi<=0&&smj<=0&&smk<=0&&
				  spipj<=0&&spimj<=0&&smipj<=0&&smimj<=0&&spipk<=0&&spimk<=0&&smipk<=0&&smimk<=0&&spjpk<=0&&spjmk<=0&&smjpk<=0&&smjmk<=0&&
			    spipjpk<=0&&spipjmk<=0&&spimjpk<=0&&spimjmk<=0&&smipjpk<=0&&smipjmk<=0&&smimjpk<=0&&smimjmk<=0) ||
				 (at(i,j,k)> 0&&spi>0&&spj>0&&spk>0&&smi>0&&smj>0&&smk>0&&
				  spipj>0&&spimj>0&&smipj>0&&smimj>0&&spipk>0&&spimk>0&&smipk>0&&smimk>0&&spjpk>0&&spjmk>0&&smjpk>0&&smjmk>0&&
			    spipjpk>0&&spipjmk>0&&spimjpk>0&&spimjmk>0&&smipjpk>0&&smipjmk>0&&smimjpk>0&&smimjmk>0))assignnew=true;
			
			if(assignnew)heapIndex.at(i,j,k)=lsNode::extend;
			else heapIndex.at(i,j,k)=lsNode::locked;
		}
		if(doPositive)for(unsigned int n=0;n<bandPositive.size();n++){
			int i=bandPositive[n].i;
			int j=bandPositive[n].j;
			int k=bandPositive[n].k;
			bool assignnew=false;
			if(i && fabs(at(i,j,k)-at(i-1,j,k))>1.1) assignnew=true;
			else if(i<I-1 && fabs(at(i,j,k)-at(i+1,j,k))>1.1) assignnew=true;
			else if(j && fabs(at(i,j,k)-at(i,j-1,k))>1.1) assignnew=true;
			else if(j<J-1 && fabs(at(i,j,k)-at(i,j+1,k))>1.1) assignnew=true;
			else if(k && fabs(at(i,j,k)-at(i,j,k-1))>1.1) assignnew=true;
			else if(k<K-1 && fabs(at(i,j,k)-at(i,j,k+1))>1.1) assignnew=true;
			
			float spi=(i<I-1)?at(i+1,j,k):at(i,j,k);
			float spj=(j<J-1)?at(i,j+1,k):at(i,j,k);
			float spk=(k<K-1)?at(i,j,k+1):at(i,j,k);
			float smi=(i)?at(i-1,j,k):at(i,j,k);
			float smj=(j)?at(i,j-1,k):at(i,j,k);
			float smk=(k)?at(i,j,k-1):at(i,j,k);
			float spipj=(i<I-1 && j<J-1)?at(i+1,j+1,k):at(i,j,k);
			float spimj=(i<I-1 && j)?at(i+1,j-1,k):at(i,j,k);
			float smipj=(i && j<J-1)?at(i-1,j+1,k):at(i,j,k);
			float smimj=(i && j)?at(i-1,j-1,k):at(i,j,k);
			float spipk=(i<I-1 && k<K-1)?at(i+1,j,k+1):at(i,j,k);
			float spimk=(i<I-1 && k)?at(i+1,j,k-1):at(i,j,k);
			float smipk=(i && k<K-1)?at(i-1,j,k+1):at(i,j,k);
			float smimk=(i && k)?at(i-1,j,k-1):at(i,j,k);
			float spjpk=(j<J-1 && k<K-1)?at(i,j+1,k+1):at(i,j,k);
			float spjmk=(j<J-1 && k)?at(i,j+1,k-1):at(i,j,k);
			float smjpk=(j && k<K-1)?at(i,j-1,k+1):at(i,j,k);
			float smjmk=(j && k)?at(i,j-1,k-1):at(i,j,k);
			float spipjpk=(i<I-1 && j<J-1 && k<K-1)?at(i+1,j+1,k+1):at(i,j,k);
			float spipjmk=(i<I-1 && j<J-1 && k)?at(i+1,j+1,k-1):at(i,j,k);
			float spimjpk=(i<I-1 && j && k<K-1)?at(i+1,j-1,k+1):at(i,j,k);
			float spimjmk=(i<I-1 && j && k)?at(i+1,j-1,k-1):at(i,j,k);
			float smipjpk=(i && j<J-1 && k<K-1)?at(i-1,j+1,k+1):at(i,j,k);
			float smipjmk=(i && j<J-1 && k)?at(i-1,j+1,k-1):at(i,j,k);
			float smimjpk=(i && j && k<K-1)?at(i-1,j-1,k+1):at(i,j,k);
			float smimjmk=(i && j && k)?at(i-1,j-1,k-1):at(i,j,k);
			
			if((at(i,j,k)<=0&&spi<=0&&spj<=0&&spk<=0&&smi<=0&&smj<=0&&smk<=0&&
				  spipj<=0&&spimj<=0&&smipj<=0&&smimj<=0&&spipk<=0&&spimk<=0&&smipk<=0&&smimk<=0&&spjpk<=0&&spjmk<=0&&smjpk<=0&&smjmk<=0&&
			    spipjpk<=0&&spipjmk<=0&&spimjpk<=0&&spimjmk<=0&&smipjpk<=0&&smipjmk<=0&&smimjpk<=0&&smimjmk<=0) ||
				 (at(i,j,k)> 0&&spi>0&&spj>0&&spk>0&&smi>0&&smj>0&&smk>0&&
				  spipj>0&&spimj>0&&smipj>0&&smimj>0&&spipk>0&&spimk>0&&smipk>0&&smimk>0&&spjpk>0&&spjmk>0&&smjpk>0&&smjmk>0&&
			    spipjpk>0&&spipjmk>0&&spimjpk>0&&spimjmk>0&&smipjpk>0&&smipjmk>0&&smimjpk>0&&smimjmk>0))assignnew=true;
			
			if(assignnew)heapIndex.at(i,j,k)=lsNode::extend;
			else heapIndex.at(i,j,k)=lsNode::locked;
		}
	}

	//////////////////////////////////
	
	// changed by VIVEK, to avoid allocation every time step
	lsNode_heap heapPositive(heapIndex);
	lsNode_heap heapNegative(heapIndex);
	std::vector<lsNode> failedValues;
	// changed back (wasn't making too much of a difference)
	//static lsNode_heap heapPositive(heapIndex);
	//static lsNode_heap heapNegative(heapIndex);
	//static std::vector<lsNode> failedValues;
	//heapPositive.reset(heapIndex);
	//heapNegative.reset(heapIndex);
	//failedValues.resize(0);

	// Initialize the tentative bands
	for(unsigned int n=0;n<bandSurface.size();n++){
		int i=bandSurface[n].i;
		int j=bandSurface[n].j;
		int k=bandSurface[n].k;
		int ii,jj,kk;
		
    if(i){
			ii=i-1;jj=j;kk=k;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
    if(j){
			ii=i;jj=j-1;kk=k;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
    if(k){
			ii=i;jj=j;kk=k-1;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
    if(i<I-1){
			ii=i+1;jj=j;kk=k;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
    if(j<J-1){
			ii=i;jj=j+1;kk=k;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
    if(k<K-1){
			ii=i;jj=j;kk=k+1;
      if(heapIndex.at(ii,jj,kk)==lsNode::extend){
        if(at(i,j,k)>0.0){
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapPositive.push(lsNode(ii,jj,kk,tv));
					}
				} else {
					bool failed;
					float tv=TentativeFMMVal(ii,jj,kk,-1.0f,heapIndex,failed);
					if(failed){
						printf("f");fflush(0);
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						heapNegative.push(lsNode(ii,jj,kk,tv));
					}
				}
      }
    }
	}
	// now the two heaps are are initialized with possible new level set values,
	// but only the smallest (biggest for negative, but they are stored as their
	// negative, so the smallest there too) is what we need.
	// Extend in the positive direction
	// TODO Change this so it only does about a layer at a time? That way if there
	// is a "leak" in the levelset it will not "fill" with all positive values.
	for(int m=0;m<2;m++){
		lsNode_heap& curHeap=(m==1?heapNegative:heapPositive);
		float curSign=1.0;
		if(m==1){
			curSign=-1.0;
		}
	  while(!curHeap.empty()){
  	  lsNode curNode=curHeap.pop();// remember pop sets healIndex.s[i][j][k]to locked
	    int i=curNode.i;
  	  int j=curNode.j;
    	int k=curNode.k;
			at(i,j,k)=curNode.ls*curSign;// curNode.ls is stored as fabs(s[i][j][k])
			// now that the node is locked, insert its neighbor nodes if the neighbor 
			// should be extended and is not in the heap, or update its key if it is
			// already in the heap.  Indices into the heap should be >=1 because the
			// heap does not use the 0 index.
			int ii,jj,kk;
    	if(i){
				ii=i-1;jj=j;kk=k;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
	    if(j){
				ii=i;jj=j-1;kk=k;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
	    if(k){
				ii=i;jj=j;kk=k-1;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
    	if(i<I-1){
				ii=i+1;jj=j;kk=k;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
	    if(j<J-1){
				ii=i;jj=j+1;kk=k;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
	    if(k<K-1){
				ii=i;jj=j;kk=k+1;
      	if(heapIndex.at(ii,jj,kk)==lsNode::extend){
					bool failed;
					float tv = TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed);
					if(failed){
						failedValues.push_back(lsNode(ii,jj,kk,tv));
					} else {
						curHeap.push(lsNode(ii,jj,kk,tv));
					}
  	    } else if(heapIndex.at(ii,jj,kk)>0){
					bool failed;
					curHeap.change_key(ii,jj,kk,TentativeFMMVal(ii,jj,kk,curSign,heapIndex,failed));
				}
    	}
	  }
	}
	// now for all the failed values, do a gauss-Jacobi iterative scheme.
	// If a node failed to get a value, then it will be the average of it's neighbor
	// values.  If there are no neighbor values in one loop then try again with the next one.
	int countZero=10;// infinte loop safety net
	while(countZero-- && failedValues.size()){
		printf("LevelSet::RedistanceBands failedValues.size()=%lu\n",failedValues.size());
		int ii,jj,kk;
		for(unsigned int m=0;m<failedValues.size();m++){
			bool remove=false;
			int i=failedValues[m].i;
			int j=failedValues[m].j;
			int k=failedValues[m].k;
			if(heapIndex.at(i,j,k)==lsNode::locked){
				failedValues[m].ls=at(i,j,k);
				remove=true;
			} else {
				int numLockedNeighbors=0;
				REAL sumLockedNeighbors=0;
				if(i){
					ii=i-1;jj=j;kk=k;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(j){
					ii=i;jj=j-1;kk=k;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(k){
					ii=i-1;jj=j;kk=k-1;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(i<I-1){
					ii=i+1;jj=j;kk=k;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(j<J-1){
					ii=i;jj=j+1;kk=k;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(k<K-1){
					ii=i;jj=j;kk=k+1;
					if(heapIndex.at(ii,jj,kk)==lsNode::locked){
						numLockedNeighbors++;
						sumLockedNeighbors+=at(ii,jj,kk);
					}
				}
				if(numLockedNeighbors){
					failedValues[m].ls=sumLockedNeighbors/numLockedNeighbors;
					remove=true;
				}
			}
			if(remove){
				at(i,j,k)=failedValues[m].ls;
				heapIndex.at(i,j,k)=lsNode::locked;
				failedValues[m]=failedValues[failedValues.size()-1];
				m--;
				failedValues.pop_back();
			}
		}
	}
}

// /* LevelSet::RedistancePDE-Use redistanceing from Dupont/liu BFECC SL LS paper, 5.1
// //
// // let P be a phi value, (s[i][j][k])
// //  On all cells with |P| > 1.1, or D[BF][xyz](P)>1.1, or all 26 neighbors are the same sigh
// //  solve,
// //   P_t + W.grad(P)=S(P^0)
// //
// //  _t is fictious time, only need to take a few steps, get about 2 away for sure...
// //  W is S(P^0)grad(P)/|grad(P)| S is sign function (what about P=0? S=0?).
// //  in W grad(P)_x is centered if DBx and DFx have same sign... otherwise maxmod(DBx,DBy)
// //  Similarly for the y and z dimension, don't forget to normalize.
// //  maxmod is take the one with maxlen.
// //
// //  Use 3rd order WENO?  to discritize the rest?  No, use simple uw and see how far you get.
// */
// void LevelSet::RedistancePDE(int numSteps, float tau, bool haveSurface){
// 	// use s as PHI^n and the ls values stored at the bands as PHI^n+1.
// 	// Collect the sign for S(PHI^0)
// 	int b_start=0;
// 	if(haveSurface)b_start=1;
	
// 	for(int b=b_start;b<3;b++){
// 		std::vector<lsNode>& curBand=*(allBands[b]);
// 		for(unsigned int m=0;m<curBand.size();m++){
// 			int i=curBand[m].i;
// 			int j=curBand[m].j;
// 			int k=curBand[m].k;
// 			if(s[i][j][k]<0)curBand[m].sn=-1;
// 			else if(s[i][j][k]>0)curBand[m].sn=1;
// 			else curBand[m].sn=0;
// 		}
// 	}		
// 	while(numSteps-->0){
// 		for(int b=b_start;b<3;b++){
// 			std::vector<lsNode>& curBand=*(allBands[b]);
// 			for(unsigned int m=0;m<curBand.size();m++){
// 				int i=curBand[m].i;
// 				int j=curBand[m].j;
// 				int k=curBand[m].k;
// 				float PHIn=s[i][j][k];
// 				bool smooth=false;
// 				if(PHIn>1.1||PHIn<-1.1){ //A2,p.9
// 					smooth=true;
// 				} else
// 				if(fabs(DBx(i,j,k))>1.1||fabs(DBy(i,j,k))>1.1||fabs(DBz(i,j,k))>1.1||
// 					 fabs(DFx(i,j,k))>1.1||fabs(DFy(i,j,k))>1.1||fabs(DFz(i,j,k))>1.1){// A
// 					smooth=true;
// 				} else
// 				if(b){//B, know that at surface B can't hold
// 					int less=0;
// 					int grea=0;
// 					smooth=true;
// 					for(int ii=i?i-1:i;ii<=(i<I-1?i+1:i);ii++)for(int jj=j?j-1:j;jj<=(j<J-1?j+1:j);jj++)for(int kk=k?k-1:k;kk<=(k<K-1?k+1:k);kk++){
// 						if(s[ii][jj][kk]>0)grea++;
// 						else less++;
// 						if(less && grea){
// 							// no need to go through the rest if the signs differ
// 							smooth=false;
// 							break;
// 						}
// 					}
// 				}
// 				if(smooth){
// 					// Smoothing means solving P_tau + W.grad(P)=S(P^0)
// 					// P^n+1=P+tau*(S-W.grad(P))
// 					// S is stored in curBand[m].sn
// 					// Find W=S(P^0)grad(P)/|grad(P)|
// 					vec W(DCx(i,j,k),DCy(i,j,k),DCz(i,j,k));
// 					W.Normalize();
// 					W*=curBand[m].sn;
// 					// find W.grad(P)
// 					float WgradP=W.x*DUx(i,j,k,W.x)+W.y*DUy(i,j,k,W.y)+W.z*DUz(i,j,k,W.z);
// 					curBand[m].ls=PHIn+tau*(curBand[m].sn-WgradP);					
// 				} else {
// 					curBand[m].ls=PHIn;
// 				}
// 			}
// 		}
// 		for(int b=b_start;b<3;b++){
// 			std::vector<lsNode>& curBand=*(allBands[b]);
// 			for(unsigned int m=0;m<curBand.size();m++){
// 				s[curBand[m].i][curBand[m].j][curBand[m].k]=curBand[m].ls;
// 			}
// 		}		
// 	}	
// }


/* Data structure to hold an index (added by vivek) */
template <class T>
struct Index3D
{
	T i,j,k;
	Index3D() {};
	Index3D(const T& _i, const T& _j, const T& _k) : i(_i),j(_j),k(_k) {}

	bool operator<(const Index3D<T>& idx) const { return (i<idx.i || (i==idx.i && (j<idx.j || (j==idx.j && k<idx.k)))); }
	bool operator>(const Index3D<T>& idx) const { return (i>idx.i || (i==idx.i && (j>idx.j || (j==idx.j && k>idx.k)))); }
	bool operator==(const Index3D<T>& idx) const { return (i==idx.i && j==idx.j && k==idx.k); }
};

/*
template<class _Ty,
class _Ax = std::allocator<_Ty> >
class static_vector : public std::vector<_Ty,_Ax>
{
public:
	void pop_back() { if (!empty()) resize(size()-1); }
};*/

/* LevelSet::RedistanceCrudelyPastBands--give crude distance from levelset
//
//  On all cells outside the narrow band the distance will be manhatine distance
// from the band.
*/
void LevelSet::RedistanceCrudelyPastBands(){	
	// The Visited field will keep track of what nodes are already in the band
	// so that we don't add them more than once.  It starts out with a value 
	// higher than the highest band, and is marked with the band level of the
	// node that is inserted
	int BandWidth=I+J+K;
	// changed by VIVEK, to avoid allocation every time step
	//field3d<int> Visited(I,J,K,BandWidth);
	static field3d<int> Visited;
	Visited.resize(I,J,K);
	Visited = BandWidth;

	//typedef std::pair<Index3D<short>,float> IndexDist;

	static std::vector< Index3D<short> > negidx; //negative indices
	static std::vector< Index3D<short> > posidx; //positive indices
	static std::vector< Index3D<short> > tmpidx;

	negidx.resize(0);
	posidx.resize(0);
	tmpidx.resize(0);

	// All the band nodes are already done
	for(unsigned int n=0;n<bandSurface.size();n++)Visited.at(bandSurface[n].i,bandSurface[n].j,bandSurface[n].k)=0;
	for(unsigned int n=0;n<bandNegative.size();n++)Visited.at(bandNegative[n].i,bandNegative[n].j,bandNegative[n].k)=0;
	for(unsigned int n=0;n<bandPositive.size();n++)Visited.at(bandPositive[n].i,bandPositive[n].j,bandPositive[n].k)=0;

	// Set of index differences to iterate over
	/*Index3D<short> diffIdx[6];
	diffIdx[0] = Index3D<short>(-1, 0, 0);
	diffIdx[1] = Index3D<short>( 1, 0, 0);
	diffIdx[2] = Index3D<short>( 0,-1, 0);
	diffIdx[3] = Index3D<short>( 0, 1, 0);
	diffIdx[4] = Index3D<short>( 0, 0,-1);
	diffIdx[5] = Index3D<short>( 0, 0, 1);
	bool validIdx[6];
	for(int c=0;c<6; c++) {
		ii=i+diffIdx[c].i;
		jj=j+diffIdx[c].j;
		kk=k+diffIdx[c].k;
		if(!=0&&Visited.s[ii][jj][kk]>searchBand){
			Visited.s[ii][jj][kk]=searchBand;
			s[ii][jj][kk]=s[i][j][k]+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
	}*/

	/*
	// Use a priority queue to store indices. This makes sure that indices closer together in memory are accessed sequentially.
	// In the priority queue, greater<> is used because, by default it uses less<> which puts the largest element on top
	typedef Index3D<short> Index3DShort;
	typedef std::priority_queue< Index3DShort,static_vector<Index3DShort>,std::greater<Index3DShort> > IndexQueue;
	static IndexQueue posqueue;
	static IndexQueue negqueue;
	static IndexQueue tmpqueue;
	*/

	// find the first outer bands
	int i,j,k,ii,jj,kk;
	int searchBand=1;
	float scale=sqrt(3.0f)/2.0;
	float sijk;
	for(unsigned int n=0;n<bandPositive.size();n++){
		i=bandPositive[n].i;
		j=bandPositive[n].j;
		k=bandPositive[n].k;
		sijk = at(i,j,k);
		ii=i;
		jj=j;
		kk=k-1;
		if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i;
		jj=j;
		kk=k+1;
		if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i;
		jj=j-1;
		kk=k;
		if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));			
		}
		ii=i;
		jj=j+1;
		kk=k;
		if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i-1;
		jj=j;
		kk=k;
		if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i+1;
		jj=j;
		kk=k;
		if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk+scale;
			posidx.push_back(Index3D<short>(ii,jj,kk));
		}
	}
	for(unsigned int n=0;n<bandNegative.size();n++){
		i=bandNegative[n].i;
		j=bandNegative[n].j;
		k=bandNegative[n].k;
		sijk = at(i,j,k);
		ii=i;
		jj=j;
		kk=k-1;
		if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i;
		jj=j;
		kk=k+1;
		if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i;
		jj=j-1;
		kk=k;
		if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i;
		jj=j+1;
		kk=k;
		if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i-1;
		jj=j;
		kk=k;
		if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
		ii=i+1;
		jj=j;
		kk=k;
		if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
			Visited.at(ii,jj,kk)=searchBand;
			at(ii,jj,kk)=sijk-scale;
			negidx.push_back(Index3D<short>(ii,jj,kk));
		}
	}

	// ok, now we have the narrow band visited at 0, and the start of the other two at 1.
	// we will not be inserting into the bands
	std::vector< Index3D<short> >* workidx_p=&tmpidx;
	std::vector< Index3D<short> >* posidx_p=&posidx;
	std::vector< Index3D<short> >* negidx_p=&negidx;
	std::vector< Index3D<short> >* tmpidx_p;

	/*
	IndexQueue* workqueue_p=&tmpqueue;
	IndexQueue* posqueue_p=&posqueue;
	IndexQueue* negqueue_p=&negqueue;
	IndexQueue* tmpqueue_p;
	*/

	for(searchBand=2;searchBand<=BandWidth;searchBand++){
		// Search toward the positive side.
		for(unsigned int n=0;n<posidx_p->size();n++){
			const Index3D<short>& idx = (*posidx_p)[n];
			i=idx.i;
			j=idx.j;
			k=idx.k;
			sijk = at(i,j,k);
			ii=i;
			jj=j;
			kk=k-1;
			if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j;
			kk=k+1;
			if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j-1;
			kk=k;
			if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j+1;
			kk=k;
			if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i-1;
			jj=j;
			kk=k;
			if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i+1;
			jj=j;
			kk=k;
			if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk+scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
		}

		//std::cerr << "sizes: " << ip_p->size() << " " << in_p->size() << std::endl;
		//std::cerr << "capacity 1: " << ip_p->capacity() << " " << jworking_p->capacity() << std::endl;
		//std::cerr << "capacity 2: " << ip_p->capacity() << " " << jworking_p->capacity() << std::endl;
		//char temp[1024];
		//std::cin.getline(temp,1024);

		posidx_p->resize(0);
		tmpidx_p=posidx_p;
		posidx_p=workidx_p;
		workidx_p=tmpidx_p;

		/*
		tmpqueue_p=posqueue_p;
		posqueue_p=workqueue_p;
		workqueue_p=tmpqueue_p;
		*/

		// Search toward the Fluid side.
		for(unsigned int n=0;n<negidx_p->size();n++){
			const Index3D<short>& idx = (*negidx_p)[n];
			i=idx.i;
			j=idx.j;
			k=idx.k;
			sijk = at(i,j,k);
			ii=i;
			jj=j;
			kk=k-1;
			if(k!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j;
			kk=k+1;
			if(k!=(K-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j-1;
			kk=k;
			if(j!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i;
			jj=j+1;
			kk=k;
			if(j!=(J-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i-1;
			jj=j;
			kk=k;
			if(i!=0&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
			ii=i+1;
			jj=j;
			kk=k;
			if(i!=(I-1)&&Visited.at(ii,jj,kk)>searchBand){
				//Timer tmr;
				Visited.at(ii,jj,kk)=searchBand;
				at(ii,jj,kk)=sijk-scale;
				//reinitLevelSetTimer += tmr.Elapsed();
				workidx_p->push_back(Index3D<short>(ii,jj,kk));
			}
		}

		negidx_p->resize(0);
		tmpidx_p=negidx_p;
		negidx_p=workidx_p;
		workidx_p=tmpidx_p;

		/*
		tmpqueue_p=negqueue_p;
		negqueue_p=workqueue_p;
		workqueue_p=tmpqueue_p;
		*/

		if(!(posidx_p->size()+negidx_p->size()))break;// no reason to keep going if we can't
	}
}

/* LevelSet::TentativeFMMVal--Return a tentative level set value for FMM
//
// The value returned will be computed only from neighboring cells that
// have a value of lsNode::locked in indexField.
// This function is used for Reinitializing the level set with the FMM.
// This should only be run on a cell that is not locked...
// Look in OsherFedkiw 2003 level set book, sec. 7.5 for details on choosing
// candidate values.
//
// POSSIBLE BUG: check for necessary sufficient condition that P(max{phi_s})<=1?
//
// i,j,k are the indices into indexField, and indexField is assumed to be the 
// same size at the fluid sim.  direction is +1 or -1 depending on if we marching
// in the positive (air) or negative (water) direction.
// This function should always return a positive value, so be carefull when
// marching in the negative direction.
*/
float LevelSet::TentativeFMMVal(int i,int j,int k,float direction,field3d<int>& indexField,bool& failed){
	failed=false;
  float A,B,C;
  A=0;B=0;C=-1;
	bool foundLockedValue;
	float lockedValue=0;// init to get rid of compuler warning
	float minLockedVal=1e20;
	// We have to look in the 6 axis directions to get candidate values
	// If there are two in the same quadrent that are locked (i.e 
	// indexField.s[i-1][j][k]==lsNode::locked &&
	// indexField.s[i+1][j][k]==lsNode::locked), then we chose the min of the
	// two locked level set values as our term.
	// find components for x-axis
	foundLockedValue=false;
  if(i&&(indexField.at(i-1,j,k)==lsNode::locked)){
		foundLockedValue=true;
		lockedValue=direction*at(i-1,j,k);
  }
	if((i<I-1)&&(indexField.at(i+1,j,k)==lsNode::locked)){
		lockedValue=(foundLockedValue?myfmin(lockedValue,direction*at(i+1,j,k)):direction*at(i+1,j,k));
		foundLockedValue=true;
  }
	if(foundLockedValue){
		minLockedVal=lockedValue;
		A+=1.0;
		B-=2.0*lockedValue;
		C+=lockedValue*lockedValue;
	}
	// find components for y-axis
	foundLockedValue=false;
  if(j&&(indexField.at(i,j-1,k)==lsNode::locked)){
		foundLockedValue=true;
		lockedValue=direction*at(i,j-1,k);
  }
	if((j<J-1)&&(indexField.at(i,j+1,k)==lsNode::locked)){
		lockedValue=(foundLockedValue?myfmin(lockedValue,direction*at(i,j+1,k)):direction*at(i,j+1,k));
		foundLockedValue=true;
  }
	if(foundLockedValue){
		minLockedVal=myfmin(minLockedVal,lockedValue);
		A+=1.0;
		B-=2.0*lockedValue;
		C+=lockedValue*lockedValue;
	}
	// find components for z-axis
	foundLockedValue=false;
  if(k&&(indexField.at(i,j,k-1)==lsNode::locked)){
		foundLockedValue=true;
		lockedValue=direction*at(i,j,k-1);
  }
	if((k<K-1)&&(indexField.at(i,j,k+1)==lsNode::locked)){
		lockedValue=(foundLockedValue?myfmin(lockedValue,direction*at(i,j,k+1)):direction*at(i,j,k+1));
		foundLockedValue=true;
  }
	if(foundLockedValue){
		minLockedVal=myfmin(minLockedVal,lockedValue);
		A+=1.0;
		B-=2.0*lockedValue;
		C+=lockedValue*lockedValue;
	}
  if(A){
		float inSqrt=B*B-4*A*C;
		if(inSqrt<0){
			return minLockedVal+1.0;
		}
		inSqrt=sqrt(inSqrt);
	  return (-B+inSqrt)/(2.0*A);
	} else {
		failed=true;
		return(direction*at(i,j,k));
	}
}


// void LevelSet::ClampDistancePastBands(void)	// added by VIVEK
// {
// }

/****************************************************************************/

void compute_distance_field (const ScalarField<3> &raster, ScalarField<3> &dist) {
	int m = raster.m, n = raster.n, o = raster.o;
	LevelSet ls(m+2,n+2,o+2, 5);
	ls.s.assign(ls.I*ls.J*ls.K, 0.5f);
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		ls.at(i+1,j+1,k+1) = raster(i,j,k) ? -0.5f : 0.5f;
	}
	ls.Reinitialize();
	dist.resize(m+2,n+2,o+2, raster.dx);
	for (int k = 0; k < o+2; k++) for (int j = 0; j < n+2; j++) for (int i = 0; i < m+2; i++) {
		dist(i,j,k) = ls.at(i,j,k);
	}	
};
