#include "primitive.hpp"
using namespace std;

template <> void sphere_dist (float r, float dx, ScalarField<2> &dist) {
	int diam = ceil(2*r/dx);
	int m = diam+2, n = diam+2;
	dist.resize(m,n, dx);
	float center = (m-1)*0.5f;
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		dist(i,j) = sqrtf(sqf(i-center) + sqf(j-center)) - r/dx;
	}
}
template <> void sphere_dist (float r, float dx, ScalarField<3> &dist) {
	int diam = ceil(2*r/dx);
	int m = diam+2, n = diam+2, o = diam+2;
	dist.resize(m,n,o, dx);
	float center = (m-1)*0.5f;
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		dist(i,j,k) = sqrtf(sqf(i-center) + sqf(j-center) + sqf(k-center)) - r/dx;
	}
}

template <> void box_dist (const Vec2f &size, float dx, ScalarField<2> &dist) {
	int m = ceil(size[0]/dx)+2, n = ceil(size[1]/dx)+2;
	dist.resize(m,n, dx);
	Vec2f c((m-1)*0.5f, (n-1)*0.5f);
	Vec2f bmin(c - size/2/dx), bmax(c + size/2/dx);
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		bool inside = (i>bmin[0] && i<bmax[0] && j>bmin[1] && j<bmax[1]);
		if (inside) {
			float distx = fminf(i-bmin[0], bmax[0]-i);
			float disty = fminf(j-bmin[1], bmax[1]-j);
			dist(i,j) = -fminf(distx, disty);
		} else {
			float distx = fabs(i - clampf(i, bmin[0],bmax[0]));
			float disty = fabs(j - clampf(j, bmin[1],bmax[1]));
			dist(i,j) = sqrtf(sqf(distx) + sqf(disty));
		}
	}
}
template <> void box_dist (const Vec3f &size, float dx, ScalarField<3> &dist) {
	int m = ceil(size[0]/dx)+2, n = ceil(size[1]/dx)+2, o = ceil(size[2]/dx)+2;
	dist.resize(m,n,o, dx);
	Vec3f c((m-1)*0.5f, (n-1)*0.5f, (o-1)*0.5f);
	Vec3f bmin(c - size/2/dx), bmax(c + size/2/dx);
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		bool inside = (i>bmin[0] && i<bmax[0] && j>bmin[1] && j<bmax[1] && k>bmin[2] && k<bmax[2]);
		if (inside) {
			float distx = fminf(i-bmin[0], bmax[0]-i);
			float disty = fminf(j-bmin[1], bmax[1]-j);
			float distz = fminf(k-bmin[2], bmax[2]-k);
			dist(i,j,k) = -fminf(fminf(distx, disty), distz);
		} else {
			float distx = fabs(i - clampf(i, bmin[0],bmax[0]));
			float disty = fabs(j - clampf(j, bmin[1],bmax[1]));
			float distz = fabs(k - clampf(k, bmin[2],bmax[2]));
			dist(i,j,k) = sqrtf(sqf(distx) + sqf(disty) + sqf(distz));
		}
	}
}
