#ifndef __SCENEGRAPH_H__
#define __SCENEGRAPH_H__

#include <vector>
#include <string>
#include <cstring>
#include "../common.hpp"

enum NodeType {CAMERA, GRID, RIGIDBODY, LIGHT};
struct HoldBodyRender {
	/// Whether this vice-like grip is active or not
	bool active;
	/// Hold till what time
	float endtime;
	Vec3f v;
	HoldBodyRender();
};

class Node {
public:
	/// Position of center of mass
	float		location[3];
	/// Bound -> min[3], max[3] about center of mass in standard basis
	float		bound[6];
	/// Rotation info
	float		axis[3];
	float		angle;
	/// Scale info
	float		scale[3];
	NodeType	type;
	std::string	objdata;
	int		align;
	int		couple;
	int		tcouple;
	HoldBodyRender	hold;

	Node();
	Node(float _location[3], float _bound[6], NodeType _type);
	void		setRotation(float _axis[3], float _angle);
	std::string &objectdata();
	void		setLocation(float _location[3]);
};

class SceneGraph {
public:
	std::vector<Node>		scene;
	bool					load(std::string filename);
	bool					save(std::string filename);
	void					update(std::string filename, int objID);
	void					clearall();
};

#endif
