/* $Id: //depot/branches/rmanprod/rman-14.0/shaders/glass.sl#1 $  (Pixar - RenderMan Division)  $Date: 2008/09/02 $ */
/*
** Copyright (c) 1999 PIXAR.  All rights reserved.  This program or
** documentation contains proprietary confidential information and trade
** secrets of PIXAR.  Reverse engineering of object code is prohibited.
** Use of copyright notice is precautionary and does not imply
** publication.
**
**                      RESTRICTED RIGHTS NOTICE
**
** Use, duplication, or disclosure by the Government is subject to the
** following restrictions:  For civilian agencies, subparagraphs (a) through
** (d) of the Commercial Computer Software--Restricted Rights clause at
** 52.227-19 of the FAR; and, for units of the Department of Defense, DoD
** Supplement to the FAR, clause 52.227-7013 (c)(1)(ii), Rights in
** Technical Data and Computer Software.
**
** Pixar
** 1001 West Cutting Blvd.
** Richmond, CA  94804
*/
/*----------------------------------------------------------------------------
 * glass - a transparent glass shader - no refraction
 *	works for colored glass (set surface color) and clear glass
 *	(Cs = [1 1 1]).
 *
 * Ka, Ks, Kd, roughness, specularcolor - the usual meaning
 * Kr - strength of environment map reflection
 * envname - environment map name
 *---------------------------------------------------------------------------*/
surface
glasscheap( float Ka=0, Kd=0, Ks=.6, Kr=0.5, roughness=.025;
	color specularcolor=1; string envname=""; float doreflections = 1; output varying color Crefl = 0; )
{
	normal Nn = normalize(N);
	vector In = normalize(I);
	normal Nf = faceforward(Nn, In);
	vector V = -In;   /* view direction */
	vector reflDir, refrDir;
	float ior = 1.5;
	float eta = (In.Nn < 0) ? 1/ior : ior;   /* relative index of refraction */
	float kr, kt;
	float raydepth, s;
	float samples = 1;

	rayinfo("depth", raydepth);
	s = (raydepth == 0) ? samples : 0; /* only 1 ray for secondary refr/refl! */

	/* Compute kr, kt, reflDir, and refrDir.  If there is total internal
	   reflection kt is set to 0 and refrDir is set to (0,0,0). */
	fresnel(In, Nf, eta, kr, kt, reflDir, refrDir);
	kt = 1 - kr;

	float Kl;
	vector Rray;
	color highlight;

	/* N and I are not automatically normalized in this implementation. */

	highlight = specularcolor*Ks*specular(Nf,V,roughness);
	Ci = Cs * (Ka * ambient() + Kd * diffuse(Nf)) + highlight;
	/* Mirror reflection */
	if (doreflections == 1 && Kr * kr > 0 && raydepth < 1) {
		Crefl = Kr * kr * environment(envname, reflDir);
		Ci += Crefl;
	}

	/*Kl = (comp(Ci,0) + comp(Ci,1) + comp(Ci,2))/3;
	Oi = Kl + (1-Cs);*/
	Oi = kr;
}
