/* I took wave's lead and renamed starfield to KMTerranbump.sl -- tal AT renderman DOT org */

/*
 * terranbump.sl - displacement for an Earth-like planet.
 *
 *
 * DESCRIPTION:
 *      When put on a sphere, makes displacements to make Earth-like
 *   mountains and oceans.  The shader works by using a variety of fractal 
 *   turbulence and mottling techniques.
 *      Note that there is a companion surface shader "terran".  You need
 *   to use exactly the same parameters for "terran" and "terranbump" to
 *   get them to work well together.
 *
 *
 * PARAMETERS:
 *    spectral_exp, lacunarity, octaves - control the fractal characteristics
 *                of the bump pattern.
 *    bump_scale - scaling of the mountains
 *    multifractal - zero uses fBm noise, nonzero uses multifractal
 *    dist_scale - scaling for multifractal distortion
 *    offset - elevation offset
 *    sea_level - obvious
 *
 *
 * HINTS:
 *       The default values for the shader assume that the planet is
 *    represented by a unit sphere.  The texture space and/or parameters
 *    to this shader will need to be altered if the size of your planet
 *    is radically different.
 *       For best results, use with the "terran" surface shader, and
 *    add a cloud layer using either "planetclouds" or "venusclouds".
 *
 *
 * AUTHOR: Ken Musgrave.
 *    Conversion to Shading Language and minor modifications by Larry Gritz.
 *
 *
 * REFERENCES:
 *
 *
 * HISTORY:
 *    ???? - original texture developed by F. Ken Musgrave.
 *    Feb 1994 - Conversion to Shading Language by L. Gritz
 *
 * last modified 1 March 1994 by lg
 */


#ifdef BMRT
#define snoise(x) (2*(noise(x)-0.5))
#else
/* This is because PRMAN's noise has less range than BMRT's */
#define snoise(x) (2.5*(noise(x)-0.5))
#endif

#define DNoise(x) ((2*(point noise(x))) - point(1,1,1))
#define VLNoise(Pt,scale) (snoise(DNoise(Pt)+(scale*Pt)))
#define N_OFFSET 0.7
#define VERY_SMALL 0.0001



displacement
tshader2 (float bump_scale = 0.04;
	  float dist_scale = .2;
	  float octaves = 4;
	    )
{
  point Ptexture;

  Ptexture = transform ("shader", P);
  float beta = bump_scale,i;
    for(i=0; i<octaves; i+=1) {
      P+=bump_scale*DNoise(pow(2,i)*Ptexture*dist_scale)/pow(2,i*beta) * normalize(Ng);
      beta*=0.4;
    }
  


  N = calculatenormal (P);
}
