/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef COUPLING_HPP
#define COUPLING_HPP

#include "rigid.hpp"
#include "sand.hpp"
#include "simhelpers.hpp"
#include "../math/sparse_matrix.h"

template <size_t d> void precompute_coupling_matrices (const Sand<d> &sand, RigidBody<d> &body);

template <size_t d> void pressure_coupling (const Sand<d> &sand, const RigidBody<d> &body, float dt, const IntArray<d> &idx, SparseMatrixf &A, std::vector<float> &b);

template <size_t d> Screw<d> pressure_force (const Sand<d> &sand, const RigidBody<d> &body);

void friction_xx_yy_coupling (const Sand<2> &sand, const RigidBody<2> &body, float dt, const IntArray<2> &idx, SparseMatrixf &A, std::vector<float> &b);
void friction_xy_coupling (const Sand<2> &sand, const RigidBody<2> &body, float dt, const IntArray<2> &idx, SparseMatrixf &A, std::vector<float> &b);

Screw<2> friction_xx_yy_force (const Sand<2> &sand, const RigidBody<2> &body);
Screw<2> friction_xy_force (const Sand<2> &sand, const RigidBody<2> &body);

void friction_xx_zz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, std::vector<float> &b);
void friction_yy_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, std::vector<float> &b);
void friction_xy_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, std::vector<float> &b);
void friction_xz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, std::vector<float> &b);
void friction_yz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, std::vector<float> &b);

Screw<3> friction_xx_zz_force (const Sand<3> &sand, const RigidBody<3> &body);
Screw<3> friction_yy_force (const Sand<3> &sand, const RigidBody<3> &body);
Screw<3> friction_xy_force (const Sand<3> &sand, const RigidBody<3> &body);
Screw<3> friction_xz_force (const Sand<3> &sand, const RigidBody<3> &body);
Screw<3> friction_yz_force (const Sand<3> &sand, const RigidBody<3> &body);

#endif
