/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef PARTICLE_HPP
#define PARTICLE_HPP

#include "../common.hpp"

template <size_t d> class Particle {
public:
	Vector<float,d> x, v;
	float m;
	Matrix<float,d,d> r2;
	bool merge;
	Particle() : merge(false) {}
};

template <size_t d> Matrix<float,d,d> advect_radii (const Matrix<float,d,d> &r2, const Matrix<float,d,d> &J);

/// Returns eigenvectors as a matrix -> column vectors
template <size_t d> Matrix <float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, Vector<float,d> &eigenVals);
template <size_t d> Matrix <float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, Vector<float,d> &eigenVals, Vector<float,d> eigenVecs[d]);
template <size_t d> Matrix <float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2);

template <size_t d> bool needSplitMerge(Particle<d> &p, bool &split, float r, float tol = 0.01f);
template <size_t d> bool needSplitOrMergeApprox(Particle<d> &p, bool &split, float r, float tol = 0.01f);
//bool needSplitMerge(Particle<2> &p, bool &split, float r, float tol = 0.01f);
template <size_t d> void splitParticle(const Particle<d> &src, Particle<d> &dest1, Particle<d> &dest2);
template <size_t d> Particle<d> mergeParticles(Particle<d> &p1, Particle<d> &p2);
//Particle<2> mergeParticles(Particle<2> &p1, Particle<2> &p2);
//Particle<3> mergeParticles(Particle<3> &p1, Particle<3> &p2);

/// Deprecated
//bool isMergeAble(const Particle<2> &p1, const Particle<2> &p2, float r);
//float ellipseVolume2(const Particle<2> &p);
//float ellipseVolume(const Particle<2> &p);
//template <size_t d> Matrix<float,d,d> radius (const Matrix<float,d,d> &r2);

#endif
