/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef SAND_HPP
#define SAND_HPP

#include "../common.hpp"
#include "../math/fields.hpp"
#include "particle.hpp"

/**
 * Split/Marge map - m decides which op
 * split: pID[0] splits to pID[{0,1}]
 * merge: pID[{0,1}] merges into pID[0]
 * MOVE: pID[0] moved to pID[1]
 */
enum modType {SPLIT, MERGE, MOVE};
struct SplitMerge {
	modType m;
	int pID[2];

	SplitMerge() {
		m = SPLIT;
	}

	SplitMerge(modType _m, int _pID[2]) : m(_m) {
		pID[0] = _pID[0];  pID[1] = _pID[1];
	}

	SplitMerge(modType _m, int pID0, int pID1) : m(_m) {
		pID[0] = pID0;  pID[1] = pID1;
	}
};

template <size_t d> class SandBase {
public:
	float dx;
	float rho_max;
	float r;
	float mu;
	std::vector< Particle<d> > particles;
	ScalarField<d> rho;
	VectorField<d> rho_s;
	VectorField<d> v;
	struct {
		ScalarField<d> lambda_sep;
		VectorField<d> dx_sep;
		ScalarField<d> f_obs;
		VectorField<d> f_obs_s, v_obs;
		ScalarField<d> p;
		TensorField<d> s;
		VectorField<d> dp, ds;
		BucketField<d> pbin;
	} aux;
	std::vector<SplitMerge> smmap;
	struct {
		bool dont_regularize;
		bool sparse_own_velocity;
		bool flip_not_pic;
		float clamp_velocity;
		float air_drag;
		Vector<float,d> air_vel;
		float cohesion;
		float cfl;
	} hacks;
	bool zhubridson;
};
template <size_t d> class Sand;
template <> class Sand<2> : public SandBase<2> {
public:
	int m, n;
	void resize (int m, int n, float dx);
};
template <> class Sand<3> : public SandBase<3> {
public:
	int m, n, o;
	void resize (int m, int n, int o, float dx);
};

template <size_t d> void advect (Sand<d> &sand, float dt);
template <size_t d> void force (Sand<d> &sand, const Vector<float,d> &gdt);
template <size_t d> void splat (Sand<d> &sand);
template <size_t d> void splitnmerge (Sand<d> &sand);
template <size_t d> int fillBuckets (Sand<d> &sand, bool fillall = true);
template <size_t d> void sim_consistency_check(Sand<d> &sand, int level = 0);

#endif
