/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef SCRIPTED_HPP
#define SCRIPTED_HPP

#include "../common.hpp"
#include "rigid.hpp"

template <size_t d> class Simulator;

template <size_t d> class Scripted {
public:
	virtual void init (const Simulator<d> &sim) = 0;
	virtual float speed (const Simulator<d> &sim) = 0;
	virtual void run (Simulator<d> &sim, float dt) = 0;
	virtual std::string type() = 0;
};

template <size_t d> class Explosion : public Scripted<d> {
public:
	float T, R, dVdt; // time, radius, volume flux
	Vector<float,d> x;
	float t; // current time
	bool blown; // true if it has gone off
	void init (const Simulator<d> &sim);
	float speed (const Simulator<d> &sim);
	void run (Simulator<d> &sim, float dt);
	std::string type();
};

template <size_t d> class HoldBody : public Scripted<d> {
public:
	RigidBody<d> *body;
	float T, t; // time to release, current time
	RigidPose<d> holdpose, releasepose;
	Screw<d> releasetwist;
	bool released;
	void init (const Simulator<d> &sim);
	float speed (const Simulator<d> &sim);
	void run (Simulator<d> &sim, float dt);
	std::string type();
};

template <size_t d> class Source : public Scripted<d> {
public:
	Vector<float,d> x0, x1; // area of source
	Vector<float,d> v; // inflow velocity
	float rho; // inflow density
	float t0, t; // time since last inserted, current time
	void init (const Simulator<d> &sim);
	float speed (const Simulator<d> &sim);
	void run (Simulator<d> &sim, float dt);
	std::string type();
};

template <size_t d> class ScriptedBody : public Scripted<d> {
public:
	RigidBody<d> *body;
	Screw<d> twist;
	void init (const Simulator<d> &sim);
	float speed (const Simulator<d> &sim);
	void run (Simulator<d> &sim, float dt);
	std::string type();
};

#endif
