/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef SIMULATOR_H
#define SIMULATOR_H

#include "../common.hpp"
#include "../math/fields.hpp"
#include "sand.hpp"
#include "rigid.hpp"

struct RigidBodyInfo {
	std::string src;
	float scale;
	float offset;
};

template <size_t d> class Scripted;

template <size_t d> class Simulator {
public:
	int frame;
	/// dt is the maximum timestep that can be taken, simtime is the current simulation time
	float dt, simtime;
	Vector<float,d> g;
	Sand<d> sand;
	btDiscreteDynamicsWorld *bt;
	std::vector< RigidBody<d> > bodies;
	std::vector<RigidBodyInfo> body_info;
	std::vector< Scripted<d>* > scripted;
	Simulator () {
		bt = NULL;
		frame = 0;
		dt = 0; simtime = 0;
	}
};

template <size_t d> void step (Simulator<d> &sim);

#endif
